# coding: utf8
from __future__ import absolute_import, unicode_literals

import random
import ujson
import itertools
import weakref
import functools
from collections import OrderedDict
from contextlib import contextmanager
from copy import copy
from thinc.neural import Model
from thinc.neural.optimizers import Adam

from .tokenizer import Tokenizer
from .vocab import Vocab
from .lemmatizer import Lemmatizer
from .pipeline import DependencyParser, Tensorizer, Tagger, EntityRecognizer
from .pipeline import SimilarityHook, TextCategorizer, SentenceSegmenter
from .pipeline import merge_noun_chunks, merge_entities
from .compat import json_dumps, izip, basestring_
from .gold import GoldParse
from .scorer import Scorer
from ._ml import link_vectors_to_models, create_default_optimizer
from .attrs import IS_STOP
from .lang.punctuation import TOKENIZER_PREFIXES, TOKENIZER_SUFFIXES
from .lang.punctuation import TOKENIZER_INFIXES
from .lang.tokenizer_exceptions import TOKEN_MATCH
from .lang.tag_map import TAG_MAP
from .lang.lex_attrs import LEX_ATTRS, is_stop
from .errors import Errors
from . import util
from . import about


class BaseDefaults(object):
    @classmethod
    def create_lemmatizer(cls, nlp=None):
        return Lemmatizer(cls.lemma_index, cls.lemma_exc, cls.lemma_rules,
                          cls.lemma_lookup)

    @classmethod
    def create_vocab(cls, nlp=None):
        lemmatizer = cls.create_lemmatizer(nlp)
        lex_attr_getters = dict(cls.lex_attr_getters)
        # This is messy, but it's the minimal working fix to Issue #639.
        lex_attr_getters[IS_STOP] = functools.partial(is_stop,
                                                      stops=cls.stop_words)
        vocab = Vocab(lex_attr_getters=lex_attr_getters, tag_map=cls.tag_map,
                      lemmatizer=lemmatizer)
        for tag_str, exc in cls.morph_rules.items():
            for orth_str, attrs in exc.items():
                vocab.morphology.add_special_case(tag_str, orth_str, attrs)
        return vocab

    @classmethod
    def create_tokenizer(cls, nlp=None):
        rules = cls.tokenizer_exceptions
        token_match = cls.token_match
        prefix_search = (util.compile_prefix_regex(cls.prefixes).search
                         if cls.prefixes else None)
        suffix_search = (util.compile_suffix_regex(cls.suffixes).search
                         if cls.suffixes else None)
        infix_finditer = (util.compile_infix_regex(cls.infixes).finditer
                          if cls.infixes else None)
        vocab = nlp.vocab if nlp is not None else cls.create_vocab(nlp)
        return Tokenizer(vocab, rules=rules,
                         prefix_search=prefix_search,
                         suffix_search=suffix_search,
                         infix_finditer=infix_finditer,
                         token_match=token_match)

    pipe_names = ['tagger', 'parser', 'ner']
    token_match = TOKEN_MATCH
    prefixes = tuple(TOKENIZER_PREFIXES)
    suffixes = tuple(TOKENIZER_SUFFIXES)
    infixes = tuple(TOKENIZER_INFIXES)
    tag_map = dict(TAG_MAP)
    tokenizer_exceptions = {}
    stop_words = set()
    lemma_rules = {}
    lemma_exc = {}
    lemma_index = {}
    lemma_lookup = {}
    morph_rules = {}
    lex_attr_getters = LEX_ATTRS
    syntax_iterators = {}


class Language(object):
    """A text-processing pipeline. Usually you'll load this once per process,
    and pass the instance around your application.

    Defaults (class): Settings, data and factory methods for creating the `nlp`
        object and processing pipeline.
    lang (unicode): Two-letter language ID, i.e. ISO code.
    """
    Defaults = BaseDefaults
    lang = None

    factories = {
        'tokenizer': lambda nlp: nlp.Defaults.create_tokenizer(nlp),
        'tensorizer': lambda nlp, **cfg: Tensorizer(nlp.vocab, **cfg),
        'tagger': lambda nlp, **cfg: Tagger(nlp.vocab, **cfg),
        'parser': lambda nlp, **cfg: DependencyParser(nlp.vocab, **cfg),
        'ner': lambda nlp, **cfg: EntityRecognizer(nlp.vocab, **cfg),
        'similarity': lambda nlp, **cfg: SimilarityHook(nlp.vocab, **cfg),
        'textcat': lambda nlp, **cfg: TextCategorizer(nlp.vocab, **cfg),
        'sbd': lambda nlp, **cfg: SentenceSegmenter(nlp.vocab, **cfg),
        'sentencizer': lambda nlp, **cfg: SentenceSegmenter(nlp.vocab, **cfg),
        'merge_noun_chunks': lambda nlp, **cfg: merge_noun_chunks,
        'merge_entities': lambda nlp, **cfg: merge_entities
    }

    def __init__(self, vocab=True, make_doc=True, max_length=10**6, meta={}, **kwargs):
        """Initialise a Language object.

        vocab (Vocab): A `Vocab` object. If `True`, a vocab is created via
            `Language.Defaults.create_vocab`.
        make_doc (callable): A function that takes text and returns a `Doc`
            object. Usually a `Tokenizer`.
        pipeline (list): A list of annotation processes or IDs of annotation,
            processes, e.g. a `Tagger` object, or `'tagger'`. IDs are looked
            up in `Language.Defaults.factories`.
        disable (list): A list of component names to exclude from the pipeline.
            The disable list has priority over the pipeline list -- if the same
            string occurs in both, the component is not loaded.
        meta (dict): Custom meta data for the Language class. Is written to by
            models to add model meta data.
        max_length (int) :
            Maximum number of characters in a single text. The current v2 models
            may run out memory on extremely long texts, due to large internal
            allocations. You should segment these texts into meaningful units,
            e.g. paragraphs, subsections etc, before passing them to spaCy.
            Default maximum length is 1,000,000 characters (1mb). As a rule of
            thumb, if all pipeline components are enabled, spaCy's default
            models currently requires roughly 1GB of temporary memory per
            100,000 characters in one text.
        RETURNS (Language): The newly constructed object.
        """
        self._meta = dict(meta)
        self._path = None
        if vocab is True:
            factory = self.Defaults.create_vocab
            vocab = factory(self, **meta.get('vocab', {}))
            if vocab.vectors.name is None:
                vocab.vectors.name = meta.get('vectors', {}).get('name')
        self.vocab = vocab
        if make_doc is True:
            factory = self.Defaults.create_tokenizer
            make_doc = factory(self, **meta.get('tokenizer', {}))
        self.tokenizer = make_doc
        self.pipeline = []
        self.max_length = max_length
        self._optimizer = None

    @property
    def path(self):
        return self._path

    @property
    def meta(self):
        self._meta.setdefault('lang', self.vocab.lang)
        self._meta.setdefault('name', 'model')
        self._meta.setdefault('version', '0.0.0')
        self._meta.setdefault('spacy_version', '>={}'.format(about.__version__))
        self._meta.setdefault('description', '')
        self._meta.setdefault('author', '')
        self._meta.setdefault('email', '')
        self._meta.setdefault('url', '')
        self._meta.setdefault('license', '')
        self._meta['vectors'] = {'width': self.vocab.vectors_length,
                                 'vectors': len(self.vocab.vectors),
                                 'keys': self.vocab.vectors.n_keys,
                                 'name': self.vocab.vectors.name}
        self._meta['pipeline'] = self.pipe_names
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value

    # Conveniences to access pipeline components
    @property
    def tensorizer(self):
        return self.get_pipe('tensorizer')

    @property
    def tagger(self):
        return self.get_pipe('tagger')

    @property
    def parser(self):
        return self.get_pipe('parser')

    @property
    def entity(self):
        return self.get_pipe('ner')

    @property
    def matcher(self):
        return self.get_pipe('matcher')

    @property
    def pipe_names(self):
        """Get names of available pipeline components.

        RETURNS (list): List of component name strings, in order.
        """
        return [pipe_name for pipe_name, _ in self.pipeline]

    def get_pipe(self, name):
        """Get a pipeline component for a given component name.

        name (unicode): Name of pipeline component to get.
        RETURNS (callable): The pipeline component.
        """
        for pipe_name, component in self.pipeline:
            if pipe_name == name:
                return component
        raise KeyError(Errors.E001.format(name=name, opts=self.pipe_names))

    def create_pipe(self, name, config=dict()):
        """Create a pipeline component from a factory.

        name (unicode): Factory name to look up in `Language.factories`.
        config (dict): Configuration parameters to initialise component.
        RETURNS (callable): Pipeline component.
        """
        if name not in self.factories:
            raise KeyError(Errors.E002.format(name=name))
        factory = self.factories[name]
        return factory(self, **config)

    def add_pipe(self, component, name=None, before=None, after=None,
                 first=None, last=None):
        """Add a component to the processing pipeline. Valid components are
        callables that take a `Doc` object, modify it and return it. Only one
        of before/after/first/last can be set. Default behaviour is "last".

        component (callable): The pipeline component.
        name (unicode): Name of pipeline component. Overwrites existing
            component.name attribute if available. If no name is set and
            the component exposes no name attribute, component.__name__ is
            used. An error is raised if a name already exists in the pipeline.
        before (unicode): Component name to insert component directly before.
        after (unicode): Component name to insert component directly after.
        first (bool): Insert component first / not first in the pipeline.
        last (bool): Insert component last / not last in the pipeline.

        EXAMPLE:
            >>> nlp.add_pipe(component, before='ner')
            >>> nlp.add_pipe(component, name='custom_name', last=True)
        """
        if not hasattr(component, '__call__'):
            msg = Errors.E003.format(component=repr(component), name=name)
            if isinstance(component, basestring_) and component in self.factories:
                msg += Errors.E004.format(component=component)
            raise ValueError(msg)
        if name is None:
            if hasattr(component, 'name'):
                name = component.name
            elif hasattr(component, '__name__'):
                name = component.__name__
            elif (hasattr(component, '__class__') and
                  hasattr(component.__class__, '__name__')):
                name = component.__class__.__name__
            else:
                name = repr(component)
        if name in self.pipe_names:
            raise ValueError(Errors.E007.format(name=name, opts=self.pipe_names))
        if sum([bool(before), bool(after), bool(first), bool(last)]) >= 2:
            raise ValueError(Errors.E006)
        pipe = (name, component)
        if last or not any([first, before, after]):
            self.pipeline.append(pipe)
        elif first:
            self.pipeline.insert(0, pipe)
        elif before and before in self.pipe_names:
            self.pipeline.insert(self.pipe_names.index(before), pipe)
        elif after and after in self.pipe_names:
            self.pipeline.insert(self.pipe_names.index(after) + 1, pipe)
        else:
            raise ValueError(Errors.E001.format(name=before or after,
                                                opts=self.pipe_names))

    def has_pipe(self, name):
        """Check if a component name is present in the pipeline. Equivalent to
        `name in nlp.pipe_names`.

        name (unicode): Name of the component.
        RETURNS (bool): Whether a component of the name exists in the pipeline.
        """
        return name in self.pipe_names

    def replace_pipe(self, name, component):
        """Replace a component in the pipeline.

        name (unicode): Name of the component to replace.
        component (callable): Pipeline component.
        """
        if name not in self.pipe_names:
            raise ValueError(Errors.E001.format(name=name, opts=self.pipe_names))
        self.pipeline[self.pipe_names.index(name)] = (name, component)

    def rename_pipe(self, old_name, new_name):
        """Rename a pipeline component.

        old_name (unicode): Name of the component to rename.
        new_name (unicode): New name of the component.
        """
        if old_name not in self.pipe_names:
            raise ValueError(Errors.E001.format(name=old_name, opts=self.pipe_names))
        if new_name in self.pipe_names:
            raise ValueError(Errors.E007.format(name=new_name, opts=self.pipe_names))
        i = self.pipe_names.index(old_name)
        self.pipeline[i] = (new_name, self.pipeline[i][1])

    def remove_pipe(self, name):
        """Remove a component from the pipeline.

        name (unicode): Name of the component to remove.
        RETURNS (tuple): A `(name, component)` tuple of the removed component.
        """
        if name not in self.pipe_names:
            raise ValueError(Errors.E001.format(name=name, opts=self.pipe_names))
        return self.pipeline.pop(self.pipe_names.index(name))

    def __call__(self, text, disable=[]):
        """Apply the pipeline to some text. The text can span multiple sentences,
        and can contain arbtrary whitespace. Alignment into the original string
        is preserved.

        text (unicode): The text to be processed.
        disable (list): Names of the pipeline components to disable.
        RETURNS (Doc): A container for accessing the annotations.

        EXAMPLE:
            >>> tokens = nlp('An example sentence. Another example sentence.')
            >>> tokens[0].text, tokens[0].head.tag_
            ('An', 'NN')
        """
        if len(text) >= self.max_length:
            raise ValueError(Errors.E088.format(length=len(text),
                                                max_length=self.max_length))
        doc = self.make_doc(text)
        for name, proc in self.pipeline:
            if name in disable:
                continue
            if not hasattr(proc, '__call__'):
                raise ValueError(Errors.E003.format(component=type(proc), name=name))
            doc = proc(doc)
            if doc is None:
                raise ValueError(Errors.E005.format(name=name))
        return doc

    def disable_pipes(self, *names):
        """Disable one or more pipeline components. If used as a context
        manager, the pipeline will be restored to the initial state at the end
        of the block. Otherwise, a DisabledPipes object is returned, that has
        a `.restore()` method you can use to undo your changes.

        EXAMPLE:
            >>> nlp.add_pipe('parser')
            >>> nlp.add_pipe('tagger')
            >>> with nlp.disable_pipes('parser', 'tagger'):
            >>>     assert not nlp.has_pipe('parser')
            >>> assert nlp.has_pipe('parser')
            >>> disabled = nlp.disable_pipes('parser')
            >>> assert len(disabled) == 1
            >>> assert not nlp.has_pipe('parser')
            >>> disabled.restore()
            >>> assert nlp.has_pipe('parser')
        """
        return DisabledPipes(self, *names)

    def make_doc(self, text):
        return self.tokenizer(text)

    def update(self, docs, golds, drop=0., sgd=None, losses=None):
        """Update the models in the pipeline.

        docs (iterable): A batch of `Doc` objects.
        golds (iterable): A batch of `GoldParse` objects.
        drop (float): The droput rate.
        sgd (callable): An optimizer.
        RETURNS (dict): Results from the update.

        EXAMPLE:
            >>> with nlp.begin_training(gold) as (trainer, optimizer):
            >>>    for epoch in trainer.epochs(gold):
            >>>        for docs, golds in epoch:
            >>>            state = nlp.update(docs, golds, sgd=optimizer)
        """
        if len(docs) != len(golds):
            raise IndexError(Errors.E009.format(n_docs=len(docs), n_golds=len(golds)))
        if len(docs) == 0:
            return
        if sgd is None:
            if self._optimizer is None:
                self._optimizer = create_default_optimizer(Model.ops)
            sgd = self._optimizer

        # Allow dict of args to GoldParse, instead of GoldParse objects.
        gold_objs = []
        doc_objs = []
        for doc, gold in zip(docs, golds):
            if isinstance(doc, basestring_):
                doc = self.make_doc(doc)
            if not isinstance(gold, GoldParse):
                gold = GoldParse(doc, **gold)
            doc_objs.append(doc)
            gold_objs.append(gold)
        golds = gold_objs
        docs = doc_objs
        grads = {}

        def get_grads(W, dW, key=None):
            grads[key] = (W, dW)

        pipes = list(self.pipeline)
        random.shuffle(pipes)
        for name, proc in pipes:
            if not hasattr(proc, 'update'):
                continue
            grads = {}
            proc.update(docs, golds, drop=drop, sgd=get_grads, losses=losses)
            for key, (W, dW) in grads.items():
                sgd(W, dW, key=key)

    def preprocess_gold(self, docs_golds):
        """Can be called before training to pre-process gold data. By default,
        it handles nonprojectivity and adds missing tags to the tag map.

        docs_golds (iterable): Tuples of `Doc` and `GoldParse` objects.
        YIELDS (tuple): Tuples of preprocessed `Doc` and `GoldParse` objects.
        """
        for name, proc in self.pipeline:
            if hasattr(proc, 'preprocess_gold'):
                docs_golds = proc.preprocess_gold(docs_golds)
        for doc, gold in docs_golds:
            yield doc, gold

    def begin_training(self, get_gold_tuples=None, sgd=None, **cfg):
        """Allocate models, pre-process training data and acquire a trainer and
        optimizer. Used as a contextmanager.

        get_gold_tuples (function): Function returning gold data
        **cfg: Config parameters.
        RETURNS: An optimizer
        """
        if get_gold_tuples is None:
            get_gold_tuples = lambda: []
        # Populate vocab
        else:
            for _, annots_brackets in get_gold_tuples():
                for annots, _ in annots_brackets:
                    for word in annots[1]:
                        _ = self.vocab[word]
        contexts = []
        if cfg.get('device', -1) >= 0:
            device = util.use_gpu(cfg['device'])
            if self.vocab.vectors.data.shape[1] >= 1:
                self.vocab.vectors.data = Model.ops.asarray(
                    self.vocab.vectors.data)
        else:
            device = None
        link_vectors_to_models(self.vocab)
        if self.vocab.vectors.data.shape[1]:
            cfg['pretrained_vectors'] = self.vocab.vectors.name
        if sgd is None:
            sgd = create_default_optimizer(Model.ops)
        self._optimizer = sgd
        for name, proc in self.pipeline:
            if hasattr(proc, 'begin_training'):
                proc.begin_training(get_gold_tuples(),
                                    pipeline=self.pipeline,
                                    sgd=self._optimizer,
                                    **cfg)
        return self._optimizer

    def evaluate(self, docs_golds, verbose=False):
        scorer = Scorer()
        docs, golds = zip(*docs_golds)
        docs = list(docs)
        golds = list(golds)
        for name, pipe in self.pipeline:
            if not hasattr(pipe, 'pipe'):
                docs = (pipe(doc) for doc in docs)
            else:
                docs = pipe.pipe(docs, batch_size=256)
        for doc, gold in zip(docs, golds):
            if verbose:
                print(doc)
            scorer.score(doc, gold, verbose=verbose)
        return scorer

    @contextmanager
    def use_params(self, params, **cfg):
        """Replace weights of models in the pipeline with those provided in the
        params dictionary. Can be used as a contextmanager, in which case,
        models go back to their original weights after the block.

        params (dict): A dictionary of parameters keyed by model ID.
        **cfg: Config parameters.

        EXAMPLE:
            >>> with nlp.use_params(optimizer.averages):
            >>>     nlp.to_disk('/tmp/checkpoint')
        """
        contexts = [pipe.use_params(params) for name, pipe
                    in self.pipeline if hasattr(pipe, 'use_params')]
        # TODO: Having trouble with contextlib
        # Workaround: these aren't actually context managers atm.
        for context in contexts:
            try:
                next(context)
            except StopIteration:
                pass
        yield
        for context in contexts:
            try:
                next(context)
            except StopIteration:
                pass

    def pipe(self, texts, as_tuples=False, n_threads=2, batch_size=1000,
             disable=[], cleanup=False):
        """Process texts as a stream, and yield `Doc` objects in order.

        texts (iterator): A sequence of texts to process.
        as_tuples (bool):
            If set to True, inputs should be a sequence of
            (text, context) tuples. Output will then be a sequence of
            (doc, context) tuples. Defaults to False.
        n_threads (int): Currently inactive.
        batch_size (int): The number of texts to buffer.
        disable (list): Names of the pipeline components to disable.
        cleanup (bool): If True, unneeded strings are freed,
            to control memory use. Experimental.
        YIELDS (Doc): Documents in the order of the original text.

        EXAMPLE:
            >>> texts = [u'One document.', u'...', u'Lots of documents']
            >>>     for doc in nlp.pipe(texts, batch_size=50, n_threads=4):
            >>>         assert doc.is_parsed
        """
        if as_tuples:
            text_context1, text_context2 = itertools.tee(texts)
            texts = (tc[0] for tc in text_context1)
            contexts = (tc[1] for tc in text_context2)
            docs = self.pipe(texts, n_threads=n_threads, batch_size=batch_size,
                             disable=disable)
            for doc, context in izip(docs, contexts):
                yield (doc, context)
            return
        docs = (self.make_doc(text) for text in texts)
        for name, proc in self.pipeline:
            if name in disable:
                continue
            if hasattr(proc, 'pipe'):
                docs = proc.pipe(docs, n_threads=n_threads,
                                 batch_size=batch_size)
            else:
                # Apply the function, but yield the doc
                docs = _pipe(proc, docs)
        # Track weakrefs of "recent" documents, so that we can see when they
        # expire from memory. When they do, we know we don't need old strings.
        # This way, we avoid maintaining an unbounded growth in string entries
        # in the string store.
        recent_refs = weakref.WeakSet()
        old_refs = weakref.WeakSet()
        # Keep track of the original string data, so that if we flush old strings,
        # we can recover the original ones. However, we only want to do this if we're
        # really adding strings, to save up-front costs.
        original_strings_data = None
        nr_seen = 0
        for doc in docs:
            yield doc
            if cleanup:
                recent_refs.add(doc)
                if nr_seen < 10000:
                    old_refs.add(doc)
                    nr_seen += 1
                elif len(old_refs) == 0:
                    old_refs, recent_refs = recent_refs, old_refs
                    if original_strings_data is None:
                        original_strings_data = list(self.vocab.strings)
                    else:
                        keys, strings = self.vocab.strings._cleanup_stale_strings(original_strings_data)
                        self.vocab._reset_cache(keys, strings)
                        self.tokenizer._reset_cache(keys)
                    nr_seen = 0

    def to_disk(self, path, disable=tuple()):
        """Save the current state to a directory.  If a model is loaded, this
        will include the model.

        path (unicode or Path): A path to a directory, which will be created if
            it doesn't exist. Paths may be strings or `Path`-like objects.
        disable (list): Names of pipeline components to disable and prevent
            from being saved.

        EXAMPLE:
            >>> nlp.to_disk('/path/to/models')
        """
        path = util.ensure_path(path)
        serializers = OrderedDict((
            ('tokenizer', lambda p: self.tokenizer.to_disk(p, vocab=False)),
            ('meta.json', lambda p: p.open('w').write(json_dumps(self.meta)))
        ))
        for name, proc in self.pipeline:
            if not hasattr(proc, 'name'):
                continue
            if name in disable:
                continue
            if not hasattr(proc, 'to_disk'):
                continue
            serializers[name] = lambda p, proc=proc: proc.to_disk(p, vocab=False)
        serializers['vocab'] = lambda p: self.vocab.to_disk(p)
        util.to_disk(path, serializers, {p: False for p in disable})

    def from_disk(self, path, disable=tuple()):
        """Loads state from a directory. Modifies the object in place and
        returns it. If the saved `Language` object contains a model, the
        model will be loaded.

        path (unicode or Path): A path to a directory. Paths may be either
            strings or `Path`-like objects.
        disable (list): Names of the pipeline components to disable.
        RETURNS (Language): The modified `Language` object.

        EXAMPLE:
            >>> from spacy.language import Language
            >>> nlp = Language().from_disk('/path/to/models')
        """
        path = util.ensure_path(path)
        deserializers = OrderedDict((
            ('meta.json', lambda p: self.meta.update(util.read_json(p))),
            ('vocab', lambda p: (
                self.vocab.from_disk(p) and _fix_pretrained_vectors_name(self))),
            ('tokenizer', lambda p: self.tokenizer.from_disk(p, vocab=False)),
        ))
        for name, proc in self.pipeline:
            if name in disable:
                continue
            if not hasattr(proc, 'to_disk'):
                continue
            deserializers[name] = lambda p, proc=proc: proc.from_disk(p, vocab=False)
        exclude = {p: False for p in disable}
        if not (path / 'vocab').exists():
            exclude['vocab'] = True
        util.from_disk(path, deserializers, exclude)
        self._path = path
        return self

    def to_bytes(self, disable=[], **exclude):
        """Serialize the current state to a binary string.

        disable (list): Nameds of pipeline components to disable and prevent
            from being serialized.
        RETURNS (bytes): The serialized form of the `Language` object.
        """
        serializers = OrderedDict((
            ('vocab', lambda: self.vocab.to_bytes()),
            ('tokenizer', lambda: self.tokenizer.to_bytes(vocab=False)),
            ('meta', lambda: json_dumps(self.meta))
        ))
        for i, (name, proc) in enumerate(self.pipeline):
            if name in disable:
                continue
            if not hasattr(proc, 'to_bytes'):
                continue
            serializers[i] = lambda proc=proc: proc.to_bytes(vocab=False)
        return util.to_bytes(serializers, exclude)

    def from_bytes(self, bytes_data, disable=[]):
        """Load state from a binary string.

        bytes_data (bytes): The data to load from.
        disable (list): Names of the pipeline components to disable.
        RETURNS (Language): The `Language` object.
        """
        deserializers = OrderedDict((
            ('meta', lambda b: self.meta.update(ujson.loads(b))),
            ('vocab', lambda b: (
                self.vocab.from_bytes(b) and _fix_pretrained_vectors_name(self))),
            ('tokenizer', lambda b: self.tokenizer.from_bytes(b, vocab=False)),
        ))
        for i, (name, proc) in enumerate(self.pipeline):
            if name in disable:
                continue
            if not hasattr(proc, 'from_bytes'):
                continue
            deserializers[i] = lambda b, proc=proc: proc.from_bytes(b, vocab=False)
        msg = util.from_bytes(bytes_data, deserializers, {})
        return self


def _fix_pretrained_vectors_name(nlp):
    # TODO: Replace this once we handle vectors consistently as static
    # data
    if 'vectors' in nlp.meta and nlp.meta['vectors'].get('name'):
        nlp.vocab.vectors.name = nlp.meta['vectors']['name']
    elif not nlp.vocab.vectors.size:
        nlp.vocab.vectors.name = None
    elif 'name' in nlp.meta and 'lang' in nlp.meta:
        vectors_name = '%s_%s.vectors' % (nlp.meta['lang'], nlp.meta['name'])
        nlp.vocab.vectors.name = vectors_name
    else:
        raise ValueError(Errors.E092)
    if nlp.vocab.vectors.size != 0:
        link_vectors_to_models(nlp.vocab)
    for name, proc in nlp.pipeline:
        if not hasattr(proc, 'cfg'):
            continue
        proc.cfg.setdefault('deprecation_fixes', {})
        proc.cfg['deprecation_fixes']['vectors_name'] = nlp.vocab.vectors.name


class DisabledPipes(list):
    """Manager for temporary pipeline disabling."""
    def __init__(self, nlp, *names):
        self.nlp = nlp
        self.names = names
        # Important! Not deep copy -- we just want the container (but we also
        # want to support people providing arbitrarily typed nlp.pipeline
        # objects.)
        self.original_pipeline = copy(nlp.pipeline)
        list.__init__(self)
        self.extend(nlp.remove_pipe(name) for name in names)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.restore()

    def restore(self):
        '''Restore the pipeline to its state when DisabledPipes was created.'''
        current, self.nlp.pipeline = self.nlp.pipeline, self.original_pipeline
        unexpected = [name for name, pipe in current
                      if not self.nlp.has_pipe(name)]
        if unexpected:
            # Don't change the pipeline if we're raising an error.
            self.nlp.pipeline = current
            raise ValueError(Errors.E008.format(names=unexpected))
        self[:] = []


def _pipe(func, docs):
    for doc in docs:
        doc = func(doc)
        yield doc



# coding: utf8
from __future__ import unicode_literals


ADJECTIVE_RULES = [
    ["οί","ός"], # καρδιακοί
    ["ές","ός"], # επιφανειακές
    ["ές","ος"], # καρδιακές
    ["ές","ύς"], # πολλές
    ["οι","ος"],
    ["αία","ος"], # ωραία
    ["ωδη","ες"], # δασώδη
    ["ώδη","ες"],
    ["ότερη","ός"],
    ["ότερος","ός"],
    ["ότεροι", "ός"],
    ["ότερων","ός"],
    ["ότερες", "ός"],
]


NOUN_RULES = [
    ["ιά","ί"], # παιδιά
    ["ια","ι"], # ποτήρια
    ["ες","α"], # κεραμίδες
    ["ές","ά"],
    ["ές","ά"],
    ["ες","α"], # εσπερινές
    ["ες","η"], # ζάχαρη
    ["ές","ή"], # φυλακές
    ["ές","ής"], # καθηγητής
    ["α","ο"], # πρόβατα
    ["α","α"], # ζήτημα
    ["ατα","α"], # στόματα
    ["άτα","άτα"], # ντομάτα
    ["άτες","άτα"], # πατάτες
    ["ία","ία"],
    ["ιά","ιά"],
    ["οί","ός"], # υπουργοί
    ["ίας","ία"], # δικτατορίας, δυσωδείας, τρομοκρατίας
    ["άτων","ατα"], # δικαιωμάτων
    ["ώπων","ωπος"], # ανθρώπων
]


VERB_RULES = [
    ["εις", "ω"],
    ["εις","ώ"],
    ["ει","ω"],
    ["ει","ώ"],
    ["ουμε","ω"],
    ["ουμε","ώ"],
    ["ούμε","ώ"], # θεώρησα
    ["ούνε","ώ"], # 
    ["ετε","ω"],
    ["ετε","ώ"],
    ["ουν","ω"],
    ["ουν","ώ"],
    ["είς","ώ"],
    ["εί","ώ"],
    ["ούν","ώ"],
    ["εσαι","ομαι"], #αισθάνεσαι
    ["εσαι","όμαι"],
    ["έσαι","ομαι"],
    ["έσαι","όμαι"],
    ["εται","ομαι"],
    ["εται","όμαι"],
    ["έται","ομαι"],
    ["έται","όμαι"],
    ["όμαστε","όμαι"],
    ["όμαστε","ομαι"],
    ["έσθε","όμαι"],
    ["εσθε","όμαι"],
    ["άς","ώ"], # αγαπάς
    ["άει","ώ"],
    ["άμε","ώ"],
    ["άτε","ώ"],
    ["άνε","ώ"],
    ["άν","ώ"],
    ["άμε","ώ"],
    ["άω","ώ"], # _verbs.py could contain any of the two
    ["ώ","άω"],
    ["όμουν", "ομαι"], # ζαλιζόμουν
    ["όμουν", "όμαι"],
    ["όμουν", "αμαι"], # κοιμόμουν
    ["όμουν", "αμαι"],
    ["ούσα", "ώ"], # ζητούσα -> ζητώ
    ["ούσες", "ώ"], 
    ["ούσε", "ώ"], 
    ["ούσαμε", "ώ"], 
    ["ούσατε", "ώ"], 
    ["ούσαν", "ώ"], 
    ["ούσανε", "ώ"], 
]


PUNCT_RULES = [
    ["“", "\""],
    ["”", "\""],
    ["\u2018", "'"],
    ["\u2019", "'"]
]



# cython: infer_types=True
# cython: profile=True
# coding: utf8
from __future__ import unicode_literals

import numpy
cimport numpy as np
import cytoolz
from collections import OrderedDict
import ujson

from .util import msgpack
from .util import msgpack_numpy

from thinc.api import chain
from thinc.v2v import Affine, SELU, Softmax
from thinc.t2v import Pooling, max_pool, mean_pool
from thinc.neural.util import to_categorical, copy_array
from thinc.neural._classes.difference import Siamese, CauchySimilarity

from .tokens.doc cimport Doc
from .syntax.nn_parser cimport Parser
from .syntax import nonproj
from .syntax.ner cimport BiluoPushDown
from .syntax.arc_eager cimport ArcEager
from .morphology cimport Morphology
from .vocab cimport Vocab
from .syntax import nonproj
from .compat import json_dumps

from .attrs import POS
from .parts_of_speech import X
from ._ml import Tok2Vec, build_text_classifier, build_tagger_model
from ._ml import link_vectors_to_models, zero_init, flatten
from ._ml import create_default_optimizer
from .errors import Errors, TempErrors
from . import util


class SentenceSegmenter(object):
    """A simple spaCy hook, to allow custom sentence boundary detection logic
    (that doesn't require the dependency parse). To change the sentence
    boundary detection strategy, pass a generator function `strategy` on
    initialization, or assign a new strategy to the .strategy attribute.
    Sentence detection strategies should be generators that take `Doc` objects
    and yield `Span` objects for each sentence.
    """
    name = 'sbd'

    def __init__(self, vocab, strategy=None):
        self.vocab = vocab
        if strategy is None or strategy == 'on_punct':
            strategy = self.split_on_punct
        self.strategy = strategy

    def __call__(self, doc):
        doc.user_hooks['sents'] = self.strategy
        return doc

    @staticmethod
    def split_on_punct(doc):
        start = 0
        seen_period = False
        for i, word in enumerate(doc):
            if seen_period and not word.is_punct:
                yield doc[start:word.i]
                start = word.i
                seen_period = False
            elif word.text in ['.', '!', '?']:
                seen_period = True
        if start < len(doc):
            yield doc[start:len(doc)]


def merge_noun_chunks(doc):
    """Merge noun chunks into a single token.

    doc (Doc): The Doc object.
    RETURNS (Doc): The Doc object with merged noun chunks.
    """
    if not doc.is_parsed:
        return doc
    spans = [(np.start_char, np.end_char, np.root.tag, np.root.dep)
             for np in doc.noun_chunks]
    for start, end, tag, dep in spans:
        doc.merge(start, end, tag=tag, dep=dep)
    return doc


def merge_entities(doc):
    """Merge entities into a single token.

    doc (Doc): The Doc object.
    RETURNS (Doc): The Doc object with merged noun entities.
    """
    spans = [(e.start_char, e.end_char, e.root.tag, e.root.dep, e.label)
             for e in doc.ents]
    for start, end, tag, dep, ent_type in spans:
        doc.merge(start, end, tag=tag, dep=dep, ent_type=ent_type)
    return doc


class Pipe(object):
    """This class is not instantiated directly. Components inherit from it, and
    it defines the interface that components should follow to function as
    components in a spaCy analysis pipeline.
    """
    name = None

    @classmethod
    def Model(cls, *shape, **kwargs):
        """Initialize a model for the pipe."""
        raise NotImplementedError

    def __init__(self, vocab, model=True, **cfg):
        """Create a new pipe instance."""
        raise NotImplementedError

    def __call__(self, doc):
        """Apply the pipe to one document. The document is
        modified in-place, and returned.

        Both __call__ and pipe should delegate to the `predict()`
        and `set_annotations()` methods.
        """
        scores, tensors = self.predict([doc])
        self.set_annotations([doc], scores, tensors=tensors)
        return doc

    def pipe(self, stream, batch_size=128, n_threads=-1):
        """Apply the pipe to a stream of documents.

        Both __call__ and pipe should delegate to the `predict()`
        and `set_annotations()` methods.
        """
        for docs in cytoolz.partition_all(batch_size, stream):
            docs = list(docs)
            scores, tensors = self.predict(docs)
            self.set_annotations(docs, scores, tensor=tensors)
            yield from docs

    def predict(self, docs):
        """Apply the pipeline's model to a batch of docs, without
        modifying them.
        """
        raise NotImplementedError

    def set_annotations(self, docs, scores, tensors=None):
        """Modify a batch of documents, using pre-computed scores."""
        raise NotImplementedError

    def update(self, docs, golds, drop=0., sgd=None, losses=None):
        """Learn from a batch of documents and gold-standard information,
        updating the pipe's model.

        Delegates to predict() and get_loss().
        """
        raise NotImplementedError

    def get_loss(self, docs, golds, scores):
        """Find the loss and gradient of loss for the batch of
        documents and their predicted scores."""
        raise NotImplementedError

    def add_label(self, label):
        """Add an output label, to be predicted by the model.

        It's possible to extend pre-trained models with new labels,
        but care should be taken to avoid the "catastrophic forgetting"
        problem.
        """
        raise NotImplementedError

    def create_optimizer(self):
        return create_default_optimizer(self.model.ops,
                                        **self.cfg.get('optimizer', {}))

    def begin_training(self, gold_tuples=tuple(), pipeline=None, sgd=None,
                       **kwargs):
        """Initialize the pipe for training, using data exampes if available.
        If no model has been initialized yet, the model is added."""
        if self.model is True:
            self.model = self.Model(**self.cfg)
        link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd

    def use_params(self, params):
        """Modify the pipe's model, to use the given parameter values."""
        with self.model.use_params(params):
            yield

    def to_bytes(self, **exclude):
        """Serialize the pipe to a bytestring."""
        serialize = OrderedDict()
        serialize['cfg'] = lambda: json_dumps(self.cfg)
        if self.model in (True, False, None):
            serialize['model'] = lambda: self.model
        else:
            serialize['model'] = self.model.to_bytes
        serialize['vocab'] = self.vocab.to_bytes
        return util.to_bytes(serialize, exclude)

    def from_bytes(self, bytes_data, **exclude):
        """Load the pipe from a bytestring."""
        def load_model(b):
            # TODO: Remove this once we don't have to handle previous models
            if self.cfg.get('pretrained_dims') and 'pretrained_vectors' not in self.cfg:
                self.cfg['pretrained_vectors'] = self.vocab.vectors.name
            if self.model is True:
                self.model = self.Model(**self.cfg)
            self.model.from_bytes(b)

        deserialize = OrderedDict((
            ('cfg', lambda b: self.cfg.update(ujson.loads(b))),
            ('vocab', lambda b: self.vocab.from_bytes(b)),
            ('model', load_model),
        ))
        util.from_bytes(bytes_data, deserialize, exclude)
        return self

    def to_disk(self, path, **exclude):
        """Serialize the pipe to disk."""
        serialize = OrderedDict()
        serialize['cfg'] = lambda p: p.open('w').write(json_dumps(self.cfg))
        serialize['vocab'] = lambda p: self.vocab.to_disk(p)
        if self.model not in (None, True, False):
            serialize['model'] = lambda p: p.open('wb').write(self.model.to_bytes())
        util.to_disk(path, serialize, exclude)

    def from_disk(self, path, **exclude):
        """Load the pipe from disk."""
        def load_model(p):
            # TODO: Remove this once we don't have to handle previous models
            if self.cfg.get('pretrained_dims') and 'pretrained_vectors' not in self.cfg:
                self.cfg['pretrained_vectors'] = self.vocab.vectors.name
            if self.model is True:
                self.model = self.Model(**self.cfg)
            self.model.from_bytes(p.open('rb').read())

        deserialize = OrderedDict((
            ('cfg', lambda p: self.cfg.update(_load_cfg(p))),
            ('vocab', lambda p: self.vocab.from_disk(p)),
            ('model', load_model),
        ))
        util.from_disk(path, deserialize, exclude)
        return self


def _load_cfg(path):
    if path.exists():
        with path.open() as file_:
            return ujson.load(file_)
    else:
        return {}


class Tensorizer(Pipe):
    """Assign position-sensitive vectors to tokens, using a CNN or RNN."""
    name = 'tensorizer'

    @classmethod
    def Model(cls, output_size=300, input_size=384, **cfg):
        """Create a new statistical model for the class.

        width (int): Output size of the model.
        embed_size (int): Number of vectors in the embedding table.
        **cfg: Config parameters.
        RETURNS (Model): A `thinc.neural.Model` or similar instance.
        """
        model = chain(
                    SELU(output_size, input_size),
                    SELU(output_size, output_size),
                    zero_init(Affine(output_size, output_size)))
        return model

    def __init__(self, vocab, model=True, **cfg):
        """Construct a new statistical model. Weights are not allocated on
        initialisation.

        vocab (Vocab): A `Vocab` instance. The model must share the same
            `Vocab` instance with the `Doc` objects it will process.
        model (Model): A `Model` instance or `True` allocate one later.
        **cfg: Config parameters.

        EXAMPLE:
            >>> from spacy.pipeline import TokenVectorEncoder
            >>> tok2vec = TokenVectorEncoder(nlp.vocab)
            >>> tok2vec.model = tok2vec.Model(128, 5000)
        """
        self.vocab = vocab
        self.model = model
        self.input_models = []
        self.cfg = dict(cfg)
        self.cfg.setdefault('cnn_maxout_pieces', 3)

    def __call__(self, doc):
        """Add context-sensitive vectors to a `Doc`, e.g. from a CNN or LSTM
        model. Vectors are set to the `Doc.tensor` attribute.

        docs (Doc or iterable): One or more documents to add vectors to.
        RETURNS (dict or None): Intermediate computations.
        """
        tokvecses = self.predict([doc])
        self.set_annotations([doc], tokvecses)
        return doc

    def pipe(self, stream, batch_size=128, n_threads=-1):
        """Process `Doc` objects as a stream.

        stream (iterator): A sequence of `Doc` objects to process.
        batch_size (int): Number of `Doc` objects to group.
        n_threads (int): Number of threads.
        YIELDS (iterator): A sequence of `Doc` objects, in order of input.
        """
        for docs in cytoolz.partition_all(batch_size, stream):
            docs = list(docs)
            tensors = self.predict(docs)
            self.set_annotations(docs, tensors)
            yield from docs

    def predict(self, docs):
        """Return a single tensor for a batch of documents.

        docs (iterable): A sequence of `Doc` objects.
        RETURNS (object): Vector representations for each token in the docs.
        """
        inputs = self.model.ops.flatten([doc.tensor for doc in docs])
        outputs = self.model(inputs)
        return self.model.ops.unflatten(outputs, [len(d) for d in docs])

    def set_annotations(self, docs, tensors):
        """Set the tensor attribute for a batch of documents.

        docs (iterable): A sequence of `Doc` objects.
        tensors (object): Vector representation for each token in the docs.
        """
        for doc, tensor in zip(docs, tensors):
            if tensor.shape[0] != len(doc):
                raise ValueError(Errors.E076.format(rows=tensor.shape[0], words=len(doc)))
            doc.tensor = tensor

    def update(self, docs, golds, state=None, drop=0., sgd=None, losses=None):
        """Update the model.

        docs (iterable): A batch of `Doc` objects.
        golds (iterable): A batch of `GoldParse` objects.
        drop (float): The droput rate.
        sgd (callable): An optimizer.
        RETURNS (dict): Results from the update.
        """
        if isinstance(docs, Doc):
            docs = [docs]
        inputs = []
        bp_inputs = []
        for tok2vec in self.input_models:
            tensor, bp_tensor = tok2vec.begin_update(docs, drop=drop)
            inputs.append(tensor)
            bp_inputs.append(bp_tensor)
        inputs = self.model.ops.xp.hstack(inputs)
        scores, bp_scores = self.model.begin_update(inputs, drop=drop)
        loss, d_scores = self.get_loss(docs, golds, scores)
        d_inputs = bp_scores(d_scores, sgd=sgd)
        d_inputs = self.model.ops.xp.split(d_inputs, len(self.input_models), axis=1)
        for d_input, bp_input in zip(d_inputs, bp_inputs):
            bp_input(d_input, sgd=sgd)
        if losses is not None:
            losses.setdefault(self.name, 0.)
            losses[self.name] += loss
        return loss

    def get_loss(self, docs, golds, prediction):
        target = []
        i = 0
        for doc in docs:
            vectors = self.model.ops.xp.vstack([w.vector for w in doc])
            target.append(vectors)
        target = self.model.ops.xp.vstack(target)
        d_scores = (prediction - target) / prediction.shape[0]
        loss = (d_scores**2).sum()
        return loss, d_scores

    def begin_training(self, gold_tuples=tuple(), pipeline=None, sgd=None,
                        **kwargs):
        """Allocate models, pre-process training data and acquire an
        optimizer.

        gold_tuples (iterable): Gold-standard training data.
        pipeline (list): The pipeline the model is part of.
        """
        for name, model in pipeline:
            if getattr(model, 'tok2vec', None):
                self.input_models.append(model.tok2vec)
        if self.model is True:
            self.cfg['input_size'] = 384
            self.cfg['output_size'] = 300
            self.model = self.Model(**self.cfg)
        link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd


class Tagger(Pipe):
    name = 'tagger'

    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.model = model
        self.cfg = OrderedDict(sorted(cfg.items()))
        self.cfg.setdefault('cnn_maxout_pieces', 2)

    @property
    def labels(self):
        return self.vocab.morphology.tag_names

    @property
    def tok2vec(self):
        if self.model in (None, True, False):
            return None
        else:
            return chain(self.model.tok2vec, flatten)

    def __call__(self, doc):
        tags, tokvecs = self.predict([doc])
        self.set_annotations([doc], tags, tensors=tokvecs)
        return doc

    def pipe(self, stream, batch_size=128, n_threads=-1):
        for docs in cytoolz.partition_all(batch_size, stream):
            docs = list(docs)
            tag_ids, tokvecs = self.predict(docs)
            self.set_annotations(docs, tag_ids, tensors=tokvecs)
            yield from docs

    def predict(self, docs):
        tokvecs = self.model.tok2vec(docs)
        scores = self.model.softmax(tokvecs)
        guesses = []
        for doc_scores in scores:
            doc_guesses = doc_scores.argmax(axis=1)
            if not isinstance(doc_guesses, numpy.ndarray):
                doc_guesses = doc_guesses.get()
            guesses.append(doc_guesses)
        return guesses, tokvecs

    def set_annotations(self, docs, batch_tag_ids, tensors=None):
        if isinstance(docs, Doc):
            docs = [docs]
        cdef Doc doc
        cdef int idx = 0
        cdef Vocab vocab = self.vocab
        for i, doc in enumerate(docs):
            doc_tag_ids = batch_tag_ids[i]
            if hasattr(doc_tag_ids, 'get'):
                doc_tag_ids = doc_tag_ids.get()
            for j, tag_id in enumerate(doc_tag_ids):
                # Don't clobber preset POS tags
                if doc.c[j].tag == 0 and doc.c[j].pos == 0:
                    # Don't clobber preset lemmas
                    lemma = doc.c[j].lemma
                    vocab.morphology.assign_tag_id(&doc.c[j], tag_id)
                    if lemma != 0 and lemma != doc.c[j].lex.orth:
                        doc.c[j].lemma = lemma
                idx += 1
            if tensors is not None:
                if isinstance(doc.tensor, numpy.ndarray) \
                and not isinstance(tensors[i], numpy.ndarray):
                    doc.extend_tensor(tensors[i].get())
                else:
                    doc.extend_tensor(tensors[i])
            doc.is_tagged = True

    def update(self, docs, golds, drop=0., sgd=None, losses=None):
        if losses is not None and self.name not in losses:
            losses[self.name] = 0.

        tag_scores, bp_tag_scores = self.model.begin_update(docs, drop=drop)
        loss, d_tag_scores = self.get_loss(docs, golds, tag_scores)
        bp_tag_scores(d_tag_scores, sgd=sgd)

        if losses is not None:
            losses[self.name] += loss

    def get_loss(self, docs, golds, scores):
        scores = self.model.ops.flatten(scores)
        tag_index = {tag: i for i, tag in enumerate(self.labels)}
        cdef int idx = 0
        correct = numpy.zeros((scores.shape[0],), dtype='i')
        guesses = scores.argmax(axis=1)
        for gold in golds:
            for tag in gold.tags:
                if tag is None:
                    correct[idx] = guesses[idx]
                else:
                    correct[idx] = tag_index[tag]
                idx += 1
        correct = self.model.ops.xp.array(correct, dtype='i')
        d_scores = scores - to_categorical(correct, nb_classes=scores.shape[1])
        d_scores /= d_scores.shape[0]
        loss = (d_scores**2).sum()
        d_scores = self.model.ops.unflatten(d_scores, [len(d) for d in docs])
        return float(loss), d_scores

    def begin_training(self, gold_tuples=tuple(), pipeline=None, sgd=None,
                       **kwargs):
        orig_tag_map = dict(self.vocab.morphology.tag_map)
        new_tag_map = OrderedDict()
        for raw_text, annots_brackets in gold_tuples:
            for annots, brackets in annots_brackets:
                ids, words, tags, heads, deps, ents = annots
                for tag in tags:
                    if tag in orig_tag_map:
                        new_tag_map[tag] = orig_tag_map[tag]
                    else:
                        new_tag_map[tag] = {POS: X}
        cdef Vocab vocab = self.vocab
        if new_tag_map:
            vocab.morphology = Morphology(vocab.strings, new_tag_map,
                                          vocab.morphology.lemmatizer,
                                          exc=vocab.morphology.exc)
        self.cfg['pretrained_vectors'] = kwargs.get('pretrained_vectors')
        if self.model is True:
            self.model = self.Model(self.vocab.morphology.n_tags, **self.cfg)
        link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd

    @classmethod
    def Model(cls, n_tags, **cfg):
        if cfg.get('pretrained_dims') and not cfg.get('pretrained_vectors'):
            raise ValueError(TempErrors.T008)
        return build_tagger_model(n_tags, **cfg)

    def add_label(self, label, values=None):
        if label in self.labels:
            return 0
        if self.model not in (True, False, None):
            # Here's how the model resizing will work, once the
            # neuron-to-tag mapping is no longer controlled by
            # the Morphology class, which sorts the tag names.
            # The sorting makes adding labels difficult.
            # smaller = self.model._layers[-1]
            # larger = Softmax(len(self.labels)+1, smaller.nI)
            # copy_array(larger.W[:smaller.nO], smaller.W)
            # copy_array(larger.b[:smaller.nO], smaller.b)
            # self.model._layers[-1] = larger
            raise ValueError(TempErrors.T003)
        tag_map = dict(self.vocab.morphology.tag_map)
        if values is None:
            values = {POS: "X"}
        tag_map[label] = values
        self.vocab.morphology = Morphology(
            self.vocab.strings, tag_map=tag_map,
            lemmatizer=self.vocab.morphology.lemmatizer,
            exc=self.vocab.morphology.exc)
        return 1

    def use_params(self, params):
        with self.model.use_params(params):
            yield

    def to_bytes(self, **exclude):
        serialize = OrderedDict()
        if self.model in (None, True, False):
            serialize['model'] = lambda: self.model
        else:
            serialize['model'] = self.model.to_bytes
        serialize['vocab'] = self.vocab.to_bytes
        serialize['cfg'] = lambda: ujson.dumps(self.cfg)
        tag_map = OrderedDict(sorted(self.vocab.morphology.tag_map.items()))
        serialize['tag_map'] = lambda: msgpack.dumps(
            tag_map, use_bin_type=True, encoding='utf8')
        return util.to_bytes(serialize, exclude)

    def from_bytes(self, bytes_data, **exclude):
        def load_model(b):
            # TODO: Remove this once we don't have to handle previous models
            if self.cfg.get('pretrained_dims') and 'pretrained_vectors' not in self.cfg:
                self.cfg['pretrained_vectors'] = self.vocab.vectors.name

            if self.model is True:
                token_vector_width = util.env_opt(
                    'token_vector_width',
                    self.cfg.get('token_vector_width', 128))
                self.model = self.Model(self.vocab.morphology.n_tags,
                                        **self.cfg)
            self.model.from_bytes(b)

        def load_tag_map(b):
            tag_map = msgpack.loads(b, encoding='utf8')
            self.vocab.morphology = Morphology(
                self.vocab.strings, tag_map=tag_map,
                lemmatizer=self.vocab.morphology.lemmatizer,
                exc=self.vocab.morphology.exc)

        deserialize = OrderedDict((
            ('vocab', lambda b: self.vocab.from_bytes(b)),
            ('tag_map', load_tag_map),
            ('cfg', lambda b: self.cfg.update(ujson.loads(b))),
            ('model', lambda b: load_model(b)),
        ))
        util.from_bytes(bytes_data, deserialize, exclude)
        return self

    def to_disk(self, path, **exclude):
        tag_map = OrderedDict(sorted(self.vocab.morphology.tag_map.items()))
        serialize = OrderedDict((
            ('vocab', lambda p: self.vocab.to_disk(p)),
            ('tag_map', lambda p: p.open('wb').write(msgpack.dumps(
                tag_map, use_bin_type=True, encoding='utf8'))),
            ('model', lambda p: p.open('wb').write(self.model.to_bytes())),
            ('cfg', lambda p: p.open('w').write(json_dumps(self.cfg)))
        ))
        util.to_disk(path, serialize, exclude)

    def from_disk(self, path, **exclude):
        def load_model(p):
            # TODO: Remove this once we don't have to handle previous models
            if self.cfg.get('pretrained_dims') and 'pretrained_vectors' not in self.cfg:
                self.cfg['pretrained_vectors'] = self.vocab.vectors.name
            if self.model is True:
                self.model = self.Model(self.vocab.morphology.n_tags, **self.cfg)
            with p.open('rb') as file_:
                self.model.from_bytes(file_.read())

        def load_tag_map(p):
            with p.open('rb') as file_:
                tag_map = msgpack.loads(file_.read(), encoding='utf8')
            self.vocab.morphology = Morphology(
                self.vocab.strings, tag_map=tag_map,
                lemmatizer=self.vocab.morphology.lemmatizer,
                exc=self.vocab.morphology.exc)

        deserialize = OrderedDict((
            ('cfg', lambda p: self.cfg.update(_load_cfg(p))),
            ('vocab', lambda p: self.vocab.from_disk(p)),
            ('tag_map', load_tag_map),
            ('model', load_model),
        ))
        util.from_disk(path, deserialize, exclude)
        return self


class MultitaskObjective(Tagger):
    """Experimental: Assist training of a parser or tagger, by training a
    side-objective.
    """
    name = 'nn_labeller'

    def __init__(self, vocab, model=True, target='dep_tag_offset', **cfg):
        self.vocab = vocab
        self.model = model
        if target == 'dep':
            self.make_label = self.make_dep
        elif target == 'tag':
            self.make_label = self.make_tag
        elif target == 'ent':
            self.make_label = self.make_ent
        elif target == 'dep_tag_offset':
            self.make_label = self.make_dep_tag_offset
        elif target == 'ent_tag':
            self.make_label = self.make_ent_tag
        elif hasattr(target, '__call__'):
            self.make_label = target
        else:
            raise ValueError(Errors.E016)
        self.cfg = dict(cfg)
        self.cfg.setdefault('cnn_maxout_pieces', 2)

    @property
    def labels(self):
        return self.cfg.setdefault('labels', {})

    @labels.setter
    def labels(self, value):
        self.cfg['labels'] = value

    def set_annotations(self, docs, dep_ids, tensors=None):
        pass

    def begin_training(self, gold_tuples=tuple(), pipeline=None, tok2vec=None,
                       sgd=None, **kwargs):
        gold_tuples = nonproj.preprocess_training_data(gold_tuples)
        for raw_text, annots_brackets in gold_tuples:
            for annots, brackets in annots_brackets:
                ids, words, tags, heads, deps, ents = annots
                for i in range(len(ids)):
                    label = self.make_label(i, words, tags, heads, deps, ents)
                    if label is not None and label not in self.labels:
                        self.labels[label] = len(self.labels)
        if self.model is True:
            token_vector_width = util.env_opt('token_vector_width')
            self.model = self.Model(len(self.labels), tok2vec=tok2vec)
        link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd

    @classmethod
    def Model(cls, n_tags, tok2vec=None, **cfg):
        token_vector_width = util.env_opt('token_vector_width', 128)
        softmax = Softmax(n_tags, token_vector_width)
        model = chain(
            tok2vec,
            softmax
        )
        model.tok2vec = tok2vec
        model.softmax = softmax
        return model

    def predict(self, docs):
        tokvecs = self.model.tok2vec(docs)
        scores = self.model.softmax(tokvecs)
        return tokvecs, scores

    def get_loss(self, docs, golds, scores):
        if len(docs) != len(golds):
            raise ValueError(Errors.E077.format(value='loss', n_docs=len(docs),
                                                n_golds=len(golds)))
        cdef int idx = 0
        correct = numpy.zeros((scores.shape[0],), dtype='i')
        guesses = scores.argmax(axis=1)
        for i, gold in enumerate(golds):
            for j in range(len(docs[i])):
                # Handes alignment for tokenization differences
                gold_idx = gold.cand_to_gold[j]
                if gold_idx is None:
                    idx += 1
                    continue
                label = self.make_label(gold_idx, gold.words, gold.tags,
                                        gold.heads, gold.labels, gold.ents)
                if label is None or label not in self.labels:
                    correct[idx] = guesses[idx]
                else:
                    correct[idx] = self.labels[label]
                idx += 1
        correct = self.model.ops.xp.array(correct, dtype='i')
        d_scores = scores - to_categorical(correct, nb_classes=scores.shape[1])
        d_scores /= d_scores.shape[0]
        loss = (d_scores**2).sum()
        return float(loss), d_scores

    @staticmethod
    def make_dep(i, words, tags, heads, deps, ents):
        if deps[i] is None or heads[i] is None:
            return None
        return deps[i]

    @staticmethod
    def make_tag(i, words, tags, heads, deps, ents):
        return tags[i]

    @staticmethod
    def make_ent(i, words, tags, heads, deps, ents):
        if ents is None:
            return None
        return ents[i]

    @staticmethod
    def make_dep_tag_offset(i, words, tags, heads, deps, ents):
        if deps[i] is None or heads[i] is None:
            return None
        offset = heads[i] - i
        offset = min(offset, 2)
        offset = max(offset, -2)
        return '%s-%s:%d' % (deps[i], tags[i], offset)

    @staticmethod
    def make_ent_tag(i, words, tags, heads, deps, ents):
        if ents is None or ents[i] is None:
            return None
        else:
            return '%s-%s' % (tags[i], ents[i])


class SimilarityHook(Pipe):
    """
    Experimental: A pipeline component to install a hook for supervised
    similarity into `Doc` objects. Requires a `Tensorizer` to pre-process
    documents. The similarity model can be any object obeying the Thinc `Model`
    interface. By default, the model concatenates the elementwise mean and
    elementwise max of the two tensors, and compares them using the
    Cauchy-like similarity function from Chen (2013):

        >>> similarity = 1. / (1. + (W * (vec1-vec2)**2).sum())

    Where W is a vector of dimension weights, initialized to 1.
    """
    name = 'similarity'

    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.model = model
        self.cfg = dict(cfg)

    @classmethod
    def Model(cls, length):
        return Siamese(Pooling(max_pool, mean_pool), CauchySimilarity(length))

    def __call__(self, doc):
        """Install similarity hook"""
        doc.user_hooks['similarity'] = self.predict
        return doc

    def pipe(self, docs, **kwargs):
        for doc in docs:
            yield self(doc)

    def predict(self, doc1, doc2):
        return self.model.predict([(doc1, doc2)])

    def update(self, doc1_doc2, golds, sgd=None, drop=0.):
        sims, bp_sims = self.model.begin_update(doc1_doc2, drop=drop)

    def begin_training(self, _=tuple(), pipeline=None, sgd=None, **kwargs):
        """Allocate model, using width from tensorizer in pipeline.

        gold_tuples (iterable): Gold-standard training data.
        pipeline (list): The pipeline the model is part of.
        """
        if self.model is True:
            self.model = self.Model(pipeline[0].model.nO)
            link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd


class TextCategorizer(Pipe):
    name = 'textcat'

    @classmethod
    def Model(cls, nr_class=1, width=64, **cfg):
        return build_text_classifier(nr_class, width, **cfg)

    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.model = model
        self.cfg = dict(cfg)

    @property
    def labels(self):
        return self.cfg.setdefault('labels', [])

    @labels.setter
    def labels(self, value):
        self.cfg['labels'] = value

    def __call__(self, doc):
        scores, tensors = self.predict([doc])
        self.set_annotations([doc], scores, tensors=tensors)
        return doc

    def pipe(self, stream, batch_size=128, n_threads=-1):
        for docs in cytoolz.partition_all(batch_size, stream):
            docs = list(docs)
            scores, tensors = self.predict(docs)
            self.set_annotations(docs, scores, tensors=tensors)
            yield from docs

    def predict(self, docs):
        scores = self.model(docs)
        scores = self.model.ops.asarray(scores)
        tensors = [doc.tensor for doc in docs]
        return scores, tensors

    def set_annotations(self, docs, scores, tensors=None):
        for i, doc in enumerate(docs):
            for j, label in enumerate(self.labels):
                doc.cats[label] = float(scores[i, j])

    def update(self, docs, golds, state=None, drop=0., sgd=None, losses=None):
        scores, bp_scores = self.model.begin_update(docs, drop=drop)
        loss, d_scores = self.get_loss(docs, golds, scores)
        bp_scores(d_scores, sgd=sgd)
        if losses is not None:
            losses.setdefault(self.name, 0.0)
            losses[self.name] += loss

    def get_loss(self, docs, golds, scores):
        truths = numpy.zeros((len(golds), len(self.labels)), dtype='f')
        not_missing = numpy.ones((len(golds), len(self.labels)), dtype='f')
        for i, gold in enumerate(golds):
            for j, label in enumerate(self.labels):
                if label in gold.cats:
                    truths[i, j] = gold.cats[label]
                else:
                    not_missing[i, j] = 0.
        truths = self.model.ops.asarray(truths)
        not_missing = self.model.ops.asarray(not_missing)
        d_scores = (scores-truths) / scores.shape[0]
        d_scores *= not_missing
        mean_square_error = ((scores-truths)**2).sum(axis=1).mean()
        return mean_square_error, d_scores

    def add_label(self, label):
        if label in self.labels:
            return 0
        if self.model not in (None, True, False):
            smaller = self.model._layers[-1]
            larger = Affine(len(self.labels)+1, smaller.nI)
            copy_array(larger.W[:smaller.nO], smaller.W)
            copy_array(larger.b[:smaller.nO], smaller.b)
            self.model._layers[-1] = larger
        self.labels.append(label)
        return 1

    def begin_training(self, gold_tuples=tuple(), pipeline=None, sgd=None,
                       **kwargs):
        if pipeline and getattr(pipeline[0], 'name', None) == 'tensorizer':
            token_vector_width = pipeline[0].model.nO
        else:
            token_vector_width = 64

        if self.model is True:
            self.cfg['pretrained_vectors'] = kwargs.get('pretrained_vectors')
            self.model = self.Model(len(self.labels), token_vector_width,
                                    **self.cfg)
            link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd


cdef class DependencyParser(Parser):
    name = 'parser'
    TransitionSystem = ArcEager

    @property
    def postprocesses(self):
        return [nonproj.deprojectivize]

    def add_multitask_objective(self, target):
        labeller = MultitaskObjective(self.vocab, target=target)
        self._multitasks.append(labeller)

    def init_multitask_objectives(self, gold_tuples, pipeline, sgd=None, **cfg):
        for labeller in self._multitasks:
            tok2vec = self.model[0]
            labeller.begin_training(gold_tuples, pipeline=pipeline,
                                    tok2vec=tok2vec, sgd=sgd)

    def __reduce__(self):
        return (DependencyParser, (self.vocab, self.moves, self.model),
                None, None)


cdef class EntityRecognizer(Parser):
    name = 'ner'
    TransitionSystem = BiluoPushDown

    nr_feature = 6

    def add_multitask_objective(self, target):
        labeller = MultitaskObjective(self.vocab, target=target)
        self._multitasks.append(labeller)

    def init_multitask_objectives(self, gold_tuples, pipeline, sgd=None, **cfg):
        for labeller in self._multitasks:
            tok2vec = self.model[0]
            labeller.begin_training(gold_tuples, pipeline=pipeline,
                                    tok2vec=tok2vec)

    def __reduce__(self):
        return (EntityRecognizer, (self.vocab, self.moves, self.model),
                None, None)


__all__ = ['Tagger', 'DependencyParser', 'EntityRecognizer', 'Tensorizer']



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

# Source https://quizlet.com/4271889/1-100-urdu-number-wordsurdu-numerals-flash-cards/
# http://www.urduword.com/lessons.php?lesson=numbers
# https://en.wikibooks.org/wiki/Urdu/Vocabulary/Numbers
# https://www.urdu-english.com/lessons/beginner/numbers

_num_words = """ایک دو تین چار پانچ چھ سات آٹھ نو دس گیارہ بارہ تیرہ چودہ پندرہ سولہ سترہ
 اٹهارا انیس بیس اکیس بائیس تئیس چوبیس پچیس چھببیس 
ستایس اٹھائس انتيس تیس اکتیس بتیس تینتیس چونتیس پینتیس
 چھتیس سینتیس ارتیس انتالیس چالیس اکتالیس بیالیس تیتالیس 
چوالیس پیتالیس چھیالیس سینتالیس اڑتالیس انچالیس پچاس اکاون باون
 تریپن چون پچپن چھپن ستاون اٹھاون انسٹھ ساثھ 
اکسٹھ باسٹھ تریسٹھ چوسٹھ پیسٹھ چھیاسٹھ سڑسٹھ اڑسٹھ 
انھتر ستر اکھتر بھتتر تیھتر چوھتر تچھتر چھیتر ستتر
اٹھتر انیاسی اسی اکیاسی بیاسی تیراسی چوراسی پچیاسی چھیاسی
 سٹیاسی اٹھیاسی نواسی نوے اکانوے بانوے ترانوے 
چورانوے پچانوے چھیانوے ستانوے اٹھانوے ننانوے سو
""".split()

# source https://www.google.com/intl/ur/inputtools/try/

_ordinal_words = """پہلا دوسرا تیسرا چوتھا پانچواں چھٹا ساتواں آٹھواں نواں دسواں گیارہواں بارہواں تیرھواں چودھواں
 پندرھواں سولہواں سترھواں اٹھارواں انیسواں بسیواں 
""".split()


def like_num(text):
    text = text.replace(',', '').replace('.', '')
    if text.isdigit():
        return True
    if text.count('/') == 1:
        num, denom = text.split('/')
        if num.isdigit() and denom.isdigit():
            return True
    if text in _num_words:
        return True
    if text in _ordinal_words:
        return True
    return False

LEX_ATTRS = {
    LIKE_NUM: like_num
}



# coding: utf8
from __future__ import unicode_literals

import pytest

DEFAULT_TESTS = [
    ('Adresa este str. Principală nr. 5.', ['Adresa', 'este', 'str.', 'Principală', 'nr.', '5', '.']),
    ('Teste, etc.', ['Teste', ',', 'etc.']),
    ('Lista, ș.a.m.d.', ['Lista', ',', 'ș.a.m.d.']),
    ('Și d.p.d.v. al...', ['Și', 'd.p.d.v.', 'al', '...'])
]

NUMBER_TESTS = [
    ('Clasa a 4-a.', ['Clasa', 'a', '4-a', '.']),
    ('Al 12-lea ceas.', ['Al', '12-lea', 'ceas', '.'])
]

TESTCASES = DEFAULT_TESTS + NUMBER_TESTS


@pytest.mark.parametrize('text,expected_tokens', TESTCASES)
def test_tokenizer_handles_testcases(ro_tokenizer, text, expected_tokens):
    tokens = ro_tokenizer(text)
    token_list = [token.text for token in tokens if not token.is_space]
    assert expected_tokens == token_list



# coding: utf8
from __future__ import unicode_literals


"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.fr.examples import sentences
>>> docs = nlp.pipe(sentences)
"""


sentences = [
    "Apple cherche a acheter une startup anglaise pour 1 milliard de dollard",
    "Les voitures autonomes voient leur assurances décalées vers les constructeurs",
    "San Francisco envisage d'interdire les robots coursiers",
    "Londres est une grande ville du Royaume-Uni",
    "L’Italie choisit ArcelorMittal pour reprendre la plus grande aciérie d’Europe",
    "Apple lance HomePod parce qu'il se sent menacé par l'Echo d'Amazon",
    "La France ne devrait pas manquer d'électricité cet été, même en cas de canicule",
    "Nouvelles attaques de Trump contre le maire de Londres",
    "Où es-tu ?",
    "Qui est le président de la France ?",
    "Où est la capitale des Etats-Unis ?",
    "Quand est né Barack Obama ?"
]



# coding: utf-8
from __future__ import unicode_literals

from ...vocab import Vocab
from ..util import get_doc


def test_issue600():
    vocab = Vocab(tag_map={'NN': {'pos': 'NOUN'}})
    doc = get_doc(vocab, ["hello"])
    doc[0].tag_ = 'NN'



# coding: utf8
from __future__ import unicode_literals

from .tokenizer_exceptions import TOKENIZER_EXCEPTIONS
from .tag_map import TAG_MAP
from .stop_words import STOP_WORDS

from ..tokenizer_exceptions import BASE_EXCEPTIONS
from ...tokens import Doc
from ..norm_exceptions import BASE_NORMS
from ...language import Language
from ...attrs import LANG, NORM
from ...util import update_exc, add_lookups


class ThaiDefaults(Language.Defaults):
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    lex_attr_getters[LANG] = lambda text: 'th'
    tokenizer_exceptions = dict(TOKENIZER_EXCEPTIONS)
    tag_map = TAG_MAP
    stop_words = STOP_WORDS


class Thai(Language):
    lang = 'th'
    Defaults = ThaiDefaults

    def make_doc(self, text):
        try:
            from pythainlp.tokenize import word_tokenize
        except ImportError:
            raise ImportError("The Thai tokenizer requires the PyThaiNLP library: "
                              "https://github.com/wannaphongcom/pythainlp/")
        words = [x for x in list(word_tokenize(text,"newmm"))]
        return Doc(self.vocab, words=words, spaces=[False]*len(words))


__all__ = ['Thai']



# encoding: utf8
from __future__ import unicode_literals

from ...symbols import ORTH, LEMMA, POS, ADV, ADJ, NOUN, ADP


_exc = {}

for exc_data in [
    {ORTH: "m.in.", LEMMA: "między innymi", POS: ADV},
    {ORTH: "inż.", LEMMA: "inżynier", POS: NOUN},
    {ORTH: "mgr.", LEMMA: "magister", POS: NOUN},
    {ORTH: "tzn.", LEMMA: "to znaczy", POS: ADV},
    {ORTH: "tj.", LEMMA: "to jest", POS: ADV},
    {ORTH: "tzw.", LEMMA: "tak zwany", POS: ADJ},
    {ORTH: "adw.", LEMMA: "adwokat", POS: NOUN},
    {ORTH: "afr.", LEMMA: "afrykański", POS: ADJ},
    {ORTH: "c.b.d.o.", LEMMA: "co było do okazania", POS: ADV},
    {ORTH: "cbdu.", LEMMA: "co było do udowodnienia", POS: ADV},
    {ORTH: "mn.w.", LEMMA: "mniej więcej", POS: ADV},
    {ORTH: "nt.", LEMMA: "na temat", POS: ADP},
    {ORTH: "ok.", LEMMA: "około"},
    {ORTH: "n.p.u.", LEMMA: "na psa urok"},
    {ORTH: "ww.", LEMMA: "wyżej wymieniony", POS: ADV}]:
    _exc[exc_data[ORTH]] = [exc_data]

for orth in [
    "w.", "r.", "br.", "bm.", "b.r.", "amer.", "am.", "bdb.", "św.", "p.", "lit.",
    "wym.", "czyt.", "daw.", "d.", "zob.", "gw.", "dn.", "dyr.", "im.", "mł.",
    "min.", "dot.", "muz.", "k.k.", "k.p.a.", "k.p.c.", "n.p.m.", "p.p.m.", "nb.",
    "ob.", "n.e.", "p.n.e.", "zw.", "zool.", "zach.", "żarg.", "żart.", "wzgl.",
    "wyj.", "xx.", "ks.", "x.", "wyd.", "wsch.", "o.o."]:
    _exc[orth] = [{ORTH: orth}]


TOKENIZER_EXCEPTIONS = _exc



'''Test Vocab.__contains__ works with int keys'''
from __future__ import unicode_literals

from ... vocab import Vocab

def test_issue1868():
    vocab = Vocab()
    lex = vocab['hello']
    assert lex.orth in vocab
    assert lex.orth_ in vocab
    assert 'some string' not in vocab
    int_id = vocab.strings.add('some string')
    assert int_id not in vocab



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

_num_words = ['нуль', 'ноль', 'бер', 'ике', 'өч', 'дүрт', 'биш', 'алты', 'җиде',
              'сигез', 'тугыз', 'ун', 'унбер', 'унике', 'унөч', 'ундүрт',
              'унбиш', 'уналты', 'унҗиде', 'унсигез', 'унтугыз', 'егерме',
              'утыз', 'кырык', 'илле', 'алтмыш', 'җитмеш', 'сиксән', 'туксан',
              'йөз', 'мең', 'төмән', 'миллион', 'миллиард', 'триллион',
              'триллиард']


def like_num(text):
    text = text.replace(',', '').replace('.', '')
    if text.isdigit():
        return True
    if text.count('/') == 1:
        num, denom = text.split('/')
        if num.isdigit() and denom.isdigit():
            return True
    if text in _num_words:
        return True
    return False


LEX_ATTRS = {
    LIKE_NUM: like_num
}



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM


_num_words = ['không', 'một', 'hai', 'ba', 'bốn', 'năm', 'sáu', 'bẩy',
              'tám', 'chín', 'mười', 'trăm', 'tỷ']


def like_num(text):
    text = text.replace(',', '').replace('.', '')
    if text.isdigit():
        return True
    if text.count('/') == 1:
        num, denom = text.split('/')
        if num.isdigit() and denom.isdigit():
            return True
    if text.lower() in _num_words:
        return True
    return False


LEX_ATTRS = {
    LIKE_NUM: like_num
}



# coding: utf8
from __future__ import unicode_literals

from .tokenizer_exceptions import TOKENIZER_EXCEPTIONS
from .norm_exceptions import NORM_EXCEPTIONS
from .tag_map import TAG_MAP
from .stop_words import STOP_WORDS
from .lex_attrs import LEX_ATTRS
from .morph_rules import MORPH_RULES
from .lemmatizer import LEMMA_RULES, LEMMA_INDEX, LEMMA_EXC, LOOKUP
from .syntax_iterators import SYNTAX_ITERATORS

from ..tokenizer_exceptions import BASE_EXCEPTIONS
from ..norm_exceptions import BASE_NORMS
from ...language import Language
from ...attrs import LANG, NORM
from ...util import update_exc, add_lookups

def _return_en(_):
    return 'en'

class EnglishDefaults(Language.Defaults):
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    lex_attr_getters.update(LEX_ATTRS)
    lex_attr_getters[LANG] = _return_en
    lex_attr_getters[NORM] = add_lookups(Language.Defaults.lex_attr_getters[NORM],
                                         BASE_NORMS, NORM_EXCEPTIONS)
    tokenizer_exceptions = update_exc(BASE_EXCEPTIONS, TOKENIZER_EXCEPTIONS)
    tag_map = TAG_MAP
    stop_words = STOP_WORDS
    morph_rules = MORPH_RULES
    lemma_rules = LEMMA_RULES
    lemma_index = LEMMA_INDEX
    lemma_exc = LEMMA_EXC
    lemma_lookup = LOOKUP
    syntax_iterators = SYNTAX_ITERATORS


class English(Language):
    lang = 'en'
    Defaults = EnglishDefaults


__all__ = ['English']



from __future__ import unicode_literals
import pytest

from ...displacy import render
from ..util import get_doc

def test_issue2361(de_tokenizer):
    tokens = de_tokenizer('< > & " ')
    html = render(get_doc(tokens.vocab, [t.text for t in tokens]))

    assert '&lt;' in html
    assert '&gt;' in html
    assert '&amp;' in html
    assert '&quot;' in html



# coding: utf8
"""  
All wordforms are extracted from Norsk Ordbank in Norwegian Bokmål 2005
(CLARINO NB - Språkbanken), Nasjonalbiblioteket, Norway:
https://www.nb.no/sprakbanken/show?serial=oai%3Anb.no%3Asbr-5&lang=en  
License:
Creative_Commons-BY (CC-BY) (https://creativecommons.org/licenses/by/4.0/)  
"""

from __future__ import unicode_literals

LOOKUP = {
'à': 'à',
'à la': 'à la',
'A/L': 'A/L',
'A/S': 'A/S',
'AD': 'AD',
'a. D': 'AD',
'a.D.': 'AD',
'aD': 'AD',
'ad': 'AD',
'ADB': 'ADB',
'adb': 'ADB',
'A.L.': 'AL',
'A/L': 'AL',
'AL': 'AL',
'a.l.': 'AL',
'a/l': 'AL',
'AM': 'AM',
'A.S': 'AS',
'A/S': 'AS',
'AS': 'AS',
'a.s.': 'AS',
'a/s': 'AS',
'as': 'AS',
'AV': 'AV',
'av': 'AV',
'Ac': 'Ac',
'Ag': 'Ag',
'Al': 'Al',
'Am': 'Am',
'Ar': 'Ar',
'As': 'As',
'At': 'At',
'Au': 'Au',
'Ba': 'Ba',
'Be': 'Be',
'Bi': 'Bi',
'Bk': 'Bk',
'Bq': 'Bq',
'Br': 'Br',
'CP': 'CP',
'cp': 'CP',
'Ca': 'Ca',
'Cd': 'Cd',
'Ce': 'Ce',
'Cf': 'Cf',
'Ci': 'Ci',
'Cl': 'Cl',
'Cm': 'Cm',
'Co': 'Co',
'Co.': 'Co.',
'co.': 'Co.',
'Cr': 'Cr',
'Cs': 'Cs',
'Cu': 'Cu',
'D/S': 'D/S',
'DK': 'DK',
'D/S': 'DS',
'DS': 'DS',
'Dy': 'Dy',
'EDB': 'EDB',
'edb': 'EDB',
'EF': 'EF',
'EKG': 'EKG',
'ekg': 'EKG',
'ES': 'ES',
'es': 'ES',
'Es': 'Es',
'Eu': 'Eu',
'EØS': 'EØS',
'eøs': 'EØS',
'FM': 'FM',
'fm': 'FM',
'FN': 'FN',
'Fe': 'Fe',
'Fm': 'Fm',
'Fr': 'Fr',
'GB': 'GB',
'GHz': 'GHz',
'GMT': 'GMT',
'gmt': 'GMT',
'GT': 'GT',
'GW': 'GW',
'Ga': 'Ga',
'Gb': 'Gb',
'Gy': 'Gy',
'H.K.H.': 'H.K.H.',
'H.k.h.': 'H.K.H.',
'H.M.': 'H.M.',
'H.m.': 'H.M.',
'HB': 'HB',
'hb': 'HB',
'HF': 'HF',
'hf': 'HF',
'HV': 'HV',
'hv': 'HV',
'He': 'He',
'Hf': 'Hf',
'Hg': 'Hg',
'Ho': 'Ho',
'Hz': 'Hz',
'Eder': 'I',
'I': 'I',
'eder': 'I',
'I': 'I',
'I. N. D.': 'I.N.D.',
'I.N.D.': 'I.N.D.',
'IQ': 'IQ',
'In': 'In',
'Ir': 'Ir',
'K/S': 'K/S',
'K/S': 'KS',
'KS': 'KS',
'k.s.': 'KS',
'k/s': 'KS',
'ks': 'KS',
'Kr': 'Kr',
'L': 'L',
'LK': 'LK',
'La': 'La',
'Lg': 'Lg',
'Li': 'Li',
'Lr': 'Lr',
'Lt.': 'Lt.',
'Lu': 'Lu',
'M': 'M',
'M/S': 'M/S',
'M/T': 'M/T',
'MB': 'MB',
'MHz': 'MHz',
'MP': 'MP',
'mp': 'MP',
'MS': 'MS',
'ms': 'MS',
'M/S': 'MS',
'MS': 'MS',
'M/T': 'MT',
'MT': 'MT',
'MV': 'MV',
'MW': 'MW',
'Mg': 'Mg',
'Mlle': 'Mlle',
'mlle': 'Mlle',
'Mme': 'Mme',
'mme': 'Mme',
'Mn': 'Mn',
'Mo': 'Mo',
'Mr.': 'Mr.',
'mr.': 'Mr.',
'Mt': 'Mt',
'NATO': 'NATO',
'Nato': 'NATO',
'NB': 'NB',
'NN': 'NN',
'NRF': 'NRF',
'NS': 'NS',
'ns': 'NS',
'Na': 'Na',
'Nato': 'Nato',
'Nb': 'Nb',
'Ne': 'Ne',
'Ng': 'Ng',
'ng': 'Ng',
'Ni': 'Ni',
'No': 'No',
'Np': 'Np',
'Oe': 'Oe',
'Os': 'Os',
'Ot.': 'Ot.',
'PVC': 'PVC',
'PVRK': 'PVRK',
'pvrk': 'PVRK',
'Pa': 'Pa',
'Pb': 'Pb',
'Pd': 'Pd',
'Pf': 'Pf',
'Pf.': 'Pf',
'Pf': 'Pf',
'Pfg': 'Pfg',
'Pfg.': 'Pfg',
'pfg': 'Pfg',
'pfg.': 'Pfg',
'Pfg': 'Pfg',
'Pm': 'Pm',
'Po': 'Po',
'Pr': 'Pr',
'Pt': 'Pt',
'Pu': 'Pu',
'Ra': 'Ra',
'Rb': 'Rb',
'Re': 'Re',
'Rh': 'Rh',
'Rn': 'Rn',
'Ru': 'Ru',
'S/S': 'S/S',
'SI': 'SI',
'S/S': 'SS',
'SS': 'SS',
'Sa': 'Sa',
'Sb': 'Sb',
'Sc': 'Sc',
'Se': 'Se',
'Sg': 'Sg',
'Si': 'Si',
'Sn': 'Sn',
'Sr': 'Sr',
'St.': 'St.',
'Sv': 'Sv',
'T/S': 'T/S',
'TCDD': 'TCDD',
'TIR': 'TIR',
'TNT': 'TNT',
'T/S': 'TS',
'TS': 'TS',
'TW': 'TW',
'Ta': 'Ta',
'Tb': 'Tb',
'Tc': 'Tc',
'Te': 'Te',
'Th': 'Th',
'Tl': 'Tl',
'Tm': 'Tm',
'V5': 'V5',
'V6': 'V6',
'VA': 'VA',
'VCM': 'VCM',
'VVS': 'VVS',
'Wb': 'Wb',
'Wh': 'Wh',
'Xe': 'Xe',
'Yb': 'Yb',
'Zn': 'Zn',
'Zr': 'Zr',
'a': 'a',
'a': 'a',
'a': 'a',
'a-': 'a-',
'a.D.': 'a.D.',
'ab': 'ab',
'ad': 'ad',
'ad': 'ad',
'adb': 'adb',
'adjø': 'adjø',
'agric.': 'agric.',
'ah': 'ah',
'aha': 'aha',
'ai': 'ai',
'akk': 'akk',
'aktenfor': 'aktenfor',
'aktenfra': 'aktenfra',
'aktenom': 'aktenom',
'akter': 'akter',
'akterinn': 'akterinn',
'akterover': 'akterover',
'akterut': 'akterut',
'all': 'all',
'alle': 'all',
'alt': 'all',
'all verden': 'all verden',
'alle vegne': 'alle vegne',
'allesteds': 'allesteds',
'allestedsfra': 'allestedsfra',
'allestedshen': 'allestedshen',
'allting': 'allting',
'alm.': 'alm.',
'amen': 'amen',
'angående': 'angående',
'andre': 'annen',
'anna': 'annen',
'annen': 'annen',
'annet': 'annen',
'annenhver': 'annenhver',
'annethvert': 'annenhver',
'annensteds': 'annensteds',
'annenstedsfra': 'annenstedsfra',
'annenstedsfra': 'annenstedsfra',
'annenstedshen': 'annenstedshen',
'annetstedsfra': 'annetstedsfra',
'art.': 'art.',
'at': 'at',
'att': 'att',
'atten': 'atten',
'attende': 'attende',
'attpå': 'attpå',
'au': 'au',
'au revoir': 'au revoir',
'aust': 'aust',
'austa': 'austa',
'austafjells': 'austafjells',
'austafor': 'austafor',
'austafra': 'austafra',
'austetter': 'austetter',
'austetter': 'austetter',
'austfra': 'austfra',
'austfra': 'austfra',
'austover': 'austover',
'austover': 'austover',
'austpå': 'austpå',
'austpå': 'austpå',
'av': 'av',
'av dage': 'av dage',
'av garde': 'av garde',
'av gårde': 'av gårde',
'av lage': 'av lage',
'av sted': 'av sted',
'av syne': 'av syne',
'babord': 'babord',
'bak': 'bak',
'bakafor': 'bakafor',
'bakenfor': 'bakenfor',
'bakenifra': 'bakenifra',
'baketter': 'baketter',
'bakfra': 'bakfra',
'bakom': 'bakom',
'bakover': 'bakover',
'bakpå': 'bakpå',
'bakut': 'bakut',
'bang': 'bang',
'basta': 'basta',
'begge': 'begge',
'beint fram': 'beint fram',
'bent fram': 'bent fram',
'bent frem': 'bent frem',
'best som': 'best som',
'best.': 'best.',
'bevares': 'bevares',
'bfr': 'bfr',
'bfr.': 'bfr',
'bl a': 'bl.a.',
'bl. a.': 'bl.a.',
'bl.a.': 'bl.a.',
'blant': 'blant',
'bnr': 'bnr.',
'bnr.': 'bnr.',
'bom': 'bom',
'bort': 'bort',
'bort imot': 'bort imot',
'bort mot': 'bort mot',
'bortafor': 'bortafor',
'borte': 'borte',
'bortefra': 'bortefra',
'bortenfor': 'bortenfor',
'borti': 'borti',
'bortigjennom': 'bortigjennom',
'bortimot': 'bortimot',
'bortom': 'bortom',
'bortover': 'bortover',
'bortpå': 'bortpå',
'borttil': 'borttil',
'bortunder': 'bortunder',
'bortved': 'bortved',
'brakar': 'brakar',
'bravissimo': 'bravissimo',
'bravo': 'bravo',
'bums': 'bums',
'bygdimellom': 'bygdimellom',
'byss': 'byss',
'både': 'både',
'bæ': 'bæ',
'bø': 'bø',
'c/o': 'c/o',
'ca': 'ca.',
'ca.': 'ca.',
'cal': 'cal',
'cand.': 'cand.',
'Cand. act.': 'cand.act.',
'cand. act': 'cand.act.',
'cand. act.': 'cand.act.',
'cand.act': 'cand.act.',
'cand.act.': 'cand.act.',
'Cand. agric.': 'cand.agric.',
'Cand.agric.': 'cand.agric.',
'cand. agric': 'cand.agric.',
'cand. agric.': 'cand.agric.',
'cand.agric': 'cand.agric.',
'cand.agric.': 'cand.agric.',
'Cand. jur.': 'cand.jur.',
'Cand.jur.': 'cand.jur.',
'cand. jur.': 'cand.jur.',
'cand.jur.': 'cand.jur.',
'Cand. Mag.': 'cand.mag.',
'Cand. mag.': 'cand.mag.',
'Cand.mag.': 'cand.mag.',
'cand. mag.': 'cand.mag.',
'cand.mag.': 'cand.mag.',
'Cand. med.': 'cand.med.',
'Cand.med.': 'cand.med.',
'cand. med': 'cand.med.',
'cand. med.': 'cand.med.',
'cand.med': 'cand.med.',
'cand.med.': 'cand.med.',
'Can.med.vet': 'cand.med.vet.',
'Cand. Med. Vet.': 'cand.med.vet.',
'Cand. med. vet': 'cand.med.vet.',
'Cand.med.vet': 'cand.med.vet.',
'Cand.med.vet.': 'cand.med.vet.',
'can.med.vet': 'cand.med.vet.',
'cand. med. vet': 'cand.med.vet.',
'cand.med.vet': 'cand.med.vet.',
'cand.med.vet.': 'cand.med.vet.',
'Cand. min.': 'cand.min.',
'Cand.min.': 'cand.min.',
'cand. min': 'cand.min.',
'cand. min.': 'cand.min.',
'cand.min': 'cand.min.',
'cand.min.': 'cand.min.',
'Cand. Odont.': 'cand.odont.',
'Cand. odont.': 'cand.odont.',
'Cand.odont.': 'cand.odont.',
'cand. odont.': 'cand.odont.',
'cand.odont.': 'cand.odont.',
'Cand. Oecon.': 'cand.oecon.',
'Cand. oecon.': 'cand.oecon.',
'Cand.oecon.': 'cand.oecon.',
'cand. oecon.': 'cand.oecon.',
'cand.oecon.': 'cand.oecon.',
'Cand. Paed.': 'cand.paed.',
'Cand. paed.': 'cand.paed.',
'Cand.paed.': 'cand.paed.',
'cand. paed.': 'cand.paed.',
'cand.paed.': 'cand.paed.',
'Cand. Pharm.': 'cand.pharm.',
'Cand. pharm.': 'cand.pharm.',
'Cand.pharm.': 'cand.pharm.',
'cand. pharm.': 'cand.pharm.',
'cand.pharm.': 'cand.pharm.',
'Cand. Philol.': 'cand.philol.',
'Cand. philol.': 'cand.philol.',
'Cand.philol.': 'cand.philol.',
'cand. philol.': 'cand.philol.',
'cand.philol.': 'cand.philol.',
'Cand. Polit.': 'cand.polit.',
'Cand. polit.': 'cand.polit.',
'Cand.polit.': 'cand.polit.',
'cand. polit.': 'cand.polit.',
'cand.polit.': 'cand.polit.',
'Cand. Psychol.': 'cand.psychol.',
'Cand. psychol.': 'cand.psychol.',
'Cand.psychol.': 'cand.psychol.',
'cand. psychol.': 'cand.psychol.',
'cand.psychol.': 'cand.psychol.',
'Cand. Real.': 'cand.real.',
'Cand. real.': 'cand.real.',
'Cand.real.': 'cand.real.',
'cand. real.': 'cand.real.',
'cand.real.': 'cand.real.',
'Cand. Scient.': 'cand.scient.',
'Cand. scient.': 'cand.scient.',
'cand. scient': 'cand.scient.',
'cand. scient.': 'cand.scient.',
'cand.scient.': 'cand.scient.',
'Cand. Sociol.': 'cand.sociol.',
'Cand. sociol.': 'cand.sociol.',
'Cand.sociol.': 'cand.sociol.',
'cand. sociol.': 'cand.sociol.',
'cand.sociol.': 'cand.sociol.',
'Cand. theol.': 'cand.theol.',
'Cand.theol.': 'cand.theol.',
'cand. theol': 'cand.theol.',
'cand. theol.': 'cand.theol.',
'cand.theol': 'cand.theol.',
'cand.theol.': 'cand.theol.',
'cd': 'cd',
'cg': 'cg',
'cl': 'cl',
'cm': 'cm',
'cm.': 'cm',
'co.': 'co.',
'contra': 'contra',
'cos': 'cos',
'cot': 'cot',
'cp': 'cp',
'd. c.': 'd.c.',
'd.c.': 'd.c.',
'd. d.': 'd.d.',
'd.d.': 'd.d.',
'd. e.': 'd.e.',
'd.e.': 'd.e.',
'd. y.': 'd.y.',
'd.y.': 'd.y.',
'dy.': 'd.y.',
'd. å.': 'd.å.',
'd.å.': 'd.å.',
'dB': 'dB',
'daa': 'daa',
'dat': 'dat.',
'dat.': 'dat.',
'datt': 'datt',
'datten': 'datten',
'db': 'db',
'de': 'de',
'dem': 'de',
'de': 'de',
'degeren': 'degeren',
'degeren': 'degeren',
'degern': 'degern',
'degern': 'degern',
'den': 'den',
'den gang': 'den gang',
'denne': 'denne',
'dept': 'dept.',
'dept.': 'dept.',
'der': 'der',
'der borte': 'der borte',
'der hen': 'der hen',
'der inne': 'der inne',
'der nede': 'der nede',
'der oppe': 'der oppe',
'der ute': 'der ute',
'dere': 'dere',
'deres': 'deres',
'derfra': 'derfra',
'deri': 'deri',
'deriblant': 'deriblant',
'derifra': 'derifra',
'derom': 'derom',
'derpå': 'derpå',
'dersom': 'dersom',
'dertil': 'dertil',
'derved': 'derved',
'det': 'det',
'det': 'det',
'det': 'det',
'dette': 'dette',
'dg': 'dg',
'dikkedikk': 'dikkedikk',
'dim.': 'dim.',
'di': 'din',
'din': 'din',
'dine': 'din',
'ditt': 'din',
'ding-dang': 'ding-dang',
'ding-ding': 'ding-ding',
'diskree': 'diskré',
'diskreere': 'diskré',
'diskreest': 'diskré',
'diskreeste': 'diskré',
'disse': 'disse',
'distree': 'distré',
'distreere': 'distré',
'distreest': 'distré',
'distreeste': 'distré',
'dit': 'dit',
'dit hen': 'dit hen',
'ditover': 'ditover',
'ditt': 'ditt',
'ditten': 'ditten',
'ditto': 'ditto',
'div': 'div.',
'div.': 'div.',
'dl': 'dl',
'dm': 'dm',
'do.': 'do.',
'Dr': 'dr.',
'Dr.': 'dr.',
'dr': 'dr.',
'dr.': 'dr.',
'Dr. agric.': 'dr.agric.',
'Dr.agric.': 'dr.agric.',
'dr. agric': 'dr.agric.',
'dr. agric.': 'dr.agric.',
'dr.agric': 'dr.agric.',
'dr.agric.': 'dr.agric.',
'Dr. Art.': 'dr.art.',
'Dr. art.': 'dr.art.',
'Dr.art.': 'dr.art.',
'dr. art.': 'dr.art.',
'dr.art.': 'dr.art.',
'Dr. ing.': 'dr.ing.',
'Dr.ing.': 'dr.ing.',
'dr. ing': 'dr.ing.',
'dr. ing.': 'dr.ing.',
'dr.ing': 'dr.ing.',
'dr.ing.': 'dr.ing.',
'Dr. Jur.': 'dr.jur.',
'Dr. jur.': 'dr.jur.',
'Dr.jur.': 'dr.jur.',
'dr.jur.': 'dr.jur.',
'Dr. med.': 'dr.med.',
'Dr.med.': 'dr.med.',
'dr. med': 'dr.med.',
'dr. med.': 'dr.med.',
'dr.med': 'dr.med.',
'dr.med.': 'dr.med.',
'Dr. Med. Vet.': 'dr.med.vet.',
'Dr. med. vet.': 'dr.med.vet.',
'Dr.med.vet.': 'dr.med.vet.',
'dr. med. vet.': 'dr.med.vet.',
'dr.med.vet.': 'dr.med.vet.',
'Dr. Odont.': 'dr.odont.',
'Dr. odont.': 'dr.odont.',
'Dr.odont.': 'dr.odont.',
'dr. odont.': 'dr.odont.',
'dr.odont.': 'dr.odont.',
'Dr. Oecon.': 'dr.oecon.',
'Dr. oecon.': 'dr.oecon.',
'Dr.oecon.': 'dr.oecon.',
'dr. oecon.': 'dr.oecon.',
'dr.oecon.': 'dr.oecon.',
'Dr. Philos.': 'dr.philos.',
'Dr. philos.': 'dr.philos.',
'Dr.philos.': 'dr.philos.',
'dr. philos.': 'dr.philos.',
'dr.philos.': 'dr.philos.',
'Dr. techn.': 'dr.techn.',
'Dr.techn.': 'dr.techn.',
'dr. techn': 'dr.techn.',
'dr. techn.': 'dr.techn.',
'dr.techn': 'dr.techn.',
'dr.techn.': 'dr.techn.',
'Dr. theol': 'dr.theol.',
'Dr.theol': 'dr.theol.',
'dr. theol': 'dr.theol.',
'dr. theol.': 'dr.theol.',
'dr.theol': 'dr.theol.',
'dr.theol.': 'dr.theol.',
'ds': 'ds.',
'ds.': 'ds.',
'dss': 'dss.',
'dss.': 'dss.',
'deg': 'du',
'du': 'du',
'dunk': 'dunk',
'duplo': 'duplo',
'dv': 'dv.',
'dv.': 'dv.',
'd. v. s.': 'dvs.',
'd.v.s.': 'dvs.',
'dvs': 'dvs.',
'dvs.': 'dvs.',
'dægeren': 'dægeren',
'dægeren': 'dægeren',
'dægern': 'dægern',
'dægern': 'dægern',
'dæven': 'dæven',
'e': 'e',
'e Kr': 'e.Kr.',
'e. Kr.': 'e.Kr.',
'e.Kr': 'e.Kr.',
'e.Kr.': 'e.Kr.',
'e f': 'e.f.',
'e. f.': 'e.f.',
'e.f.': 'e.f.',
'e l': 'e.l.',
'e. l.': 'e.l.',
'e. lign.': 'e.l.',
'e. likn.': 'e.l.',
'e.l.': 'e.l.',
'e.lign.': 'e.l.',
'e.likn.': 'e.l.',
'ell. lign.': 'e.l.',
'ell. likn.': 'e.l.',
'ell.lign.': 'e.l.',
'ell.likn.': 'e.l.',
'eV': 'eV',
'edb': 'edb',
'eflg.': 'eflg.',
'eg.': 'eg.',
'ei': 'ei',
'eia': 'eia',
'egne': 'eigen',
'eiga': 'eigen',
'eige': 'eigen',
'eigen': 'eigen',
'ekg': 'ekg',
'eksklusiv': 'eksklusiv',
'eksklusive': 'eksklusive',
'el.': 'el.',
'eleison': 'eleison',
'eller': 'eller',
'elleve': 'elleve',
'ei': 'en',
'en': 'en',
'et': 'en',
'ett': 'en',
'en': 'en',
'ei': 'en',
'en': 'en',
'et': 'en',
'ett': 'en',
'ende fram': 'ende fram',
'ende frem': 'ende frem',
'endelangs': 'endelangs',
'eneste': 'eneste',
'enhver': 'enhver',
'ethvert': 'enhver',
'enn': 'enn',
'enn': 'enn',
'enn si': 'enn si',
'enn videre': 'enn videre',
'enn ytterligere': 'enn ytterligere',
'enskjønt': 'enskjønt',
'ensteds': 'ensteds',
'enten': 'enten',
'et.': 'et.',
'etc': 'etc.',
'etc.': 'etc.',
'etsteds': 'etsteds',
'etter': 'etter',
'etter at': 'etter at',
'etter som': 'etter som',
'ettersom': 'ettersom',
'ev': 'ev.',
'ev.': 'ev.',
'evt.': 'ev.',
'ex': 'ex',
'f.': 'f.',
'f Kr': 'f.Kr.',
'f. Kr.': 'f.Kr.',
'f.Kr': 'f.Kr.',
'f.Kr.': 'f.Kr.',
'f eks': 'f.eks.',
'f. eks': 'f.eks.',
'f. eks.': 'f.eks.',
'f.eks': 'f.eks.',
'f.eks.': 'f.eks.',
'feks': 'f.eks.',
'feks.': 'f.eks.',
'f. o. r.': 'f.o.r.',
'f.o.r.': 'f.o.r.',
'face': 'face',
'fallera': 'fallera',
'falleri': 'falleri',
'far vel': 'far vel',
'farken': 'farken',
'fec.': 'fec.',
'fem': 'fem',
'femten': 'femten',
'femti': 'femti',
'ff': 'ff',
'fff': 'fff',
'fhv': 'fhv.',
'fhv.': 'fhv.',
'filiokus': 'filiokus',
'filleren': 'filleren',
'fire': 'fire',
'fjartan': 'fjartan',
'fjorten': 'fjorten',
'fl': 'fl',
'fl.': 'fl',
'fl.': 'fl.',
'fl.': 'fl.',
'flt.': 'fl.',
'flg': 'flg.',
'flg.': 'flg.',
'flt.': 'flt.',
'fol': 'fol.',
'fol.': 'fol.',
'fola': 'fola',
'for': 'for',
'for': 'for',
'for at': 'for at',
'for di': 'for di',
'for hånden': 'for hånden',
'for tida': 'for tida',
'for tiden': 'for tiden',
'foran': 'foran',
'forbi': 'forbi',
'fordi': 'fordi',
'forfra': 'forfra',
'fork': 'fork.',
'fork.': 'fork.',
'formedelst': 'formedelst',
'formødre': 'formødre',
'formødrer': 'formødrer',
'forover': 'forover',
'forrige': 'forrige',
'forut': 'forut',
'foruten': 'foruten',
'fr': 'fr',
'fr.': 'fr',
'fr': 'fr.',
'fr.': 'fr.',
'fra': 'fra',
'fra borde': 'fra borde',
'fram': 'fram',
'frametter': 'frametter',
'framfor': 'framfor',
'framføre': 'framføre',
'frammafor': 'frammafor',
'framme': 'framme',
'frammenfor': 'frammenfor',
'framom': 'framom',
'framover': 'framover',
'frampå': 'frampå',
'frem': 'frem',
'fremad': 'fremad',
'fremetter': 'fremetter',
'fremfor': 'fremfor',
'fremme': 'fremme',
'fremmenfor': 'fremmenfor',
'fremom': 'fremom',
'fremover': 'fremover',
'frempå': 'frempå',
'frk.': 'frk.',
'fut.': 'fut.',
'fy': 'fy',
'fysj': 'fysj',
'før': 'før',
'før enn': 'før enn',
'føre': 'føre',
'føre': 'føre',
'førti': 'førti',
'g': 'g.',
'g.': 'g.',
'gakk': 'gakk',
'gakk-gakk': 'gakk-gakk',
'gal': 'gal',
'gardimellom': 'gardimellom',
'gatelangs': 'gatelangs',
'gen.': 'gen.',
'gett': 'gett',
'gg': 'gg',
'gg.': 'gg',
'gid': 'gid',
'giss': 'giss',
'gitt': 'gitt',
'gjennom': 'gjennom',
'gjøren': 'gjøren',
'gl': 'gl.',
'gl.': 'gl.',
'gml': 'gl.',
'gml.': 'gl.',
'gnr': 'gnr.',
'gnr.': 'gnr.',
'god dag': 'god dag',
'god kveld': 'god kveld',
'god morgen': 'god morgen',
'god natt': 'god natt',
'gros': 'gros',
'gud skje lov': 'gud skje lov',
'går kveld': 'går kveld',
'går morges': 'går morges',
'går natt': 'går natt',
'ha': 'ha',
'ha': 'ha',
'ha-ha': 'ha-ha',
'haha': 'haha',
'halleluja': 'halleluja',
'hallo': 'hallo',
'halloi': 'halloi',
'halloi': 'halloi',
'haloi': 'haloi',
'haloi': 'haloi',
'halvanna': 'halvannen',
'halvannen': 'halvannen',
'halvannet': 'halvannen',
'halvhundre': 'halvhundre',
'halvveges': 'halvveges',
'halvvegs': 'halvvegs',
'ham': 'han',
'han': 'han',
'hands up': 'hands up',
'hans': 'hans',
'hau': 'hau',
'he': 'he',
'hei': 'hei',
'hei sann': 'hei sann',
'heia': 'heia',
'heil': 'heil',
'heim': 'heim',
'heim att': 'heim att',
'heima': 'heima',
'heimafra': 'heimafra',
'heimanfra': 'heimanfra',
'heime': 'heime',
'heimefra': 'heimefra',
'heimefra': 'heimefra',
'heimetter': 'heimetter',
'heimom': 'heimom',
'heimover': 'heimover',
'helledussen': 'helledussen',
'hen': 'hen',
'henad': 'henad',
'henimot': 'henimot',
'henne': 'henne',
'hennes': 'hennes',
'henover': 'henover',
'henved': 'henved',
'hepp': 'hepp',
'her': 'her',
'herav': 'herav',
'heretter': 'heretter',
'herfra': 'herfra',
'heri': 'heri',
'heriblant': 'heriblant',
'herifra': 'herifra',
'herigjennom': 'herigjennom',
'herimot': 'herimot',
'hermed': 'hermed',
'herom': 'herom',
'herover': 'herover',
'herpå': 'herpå',
'herregud': 'herregud',
'hertil': 'hertil',
'herunder': 'herunder',
'herved': 'herved',
'heureka': 'heureka',
'hg': 'hg',
'hhv': 'hhv.',
'hhv.': 'hhv.',
'hill': 'hill',
'hi': 'hin',
'hin': 'hin',
'hine': 'hin',
'hint': 'hin',
'hitt': 'hin',
'hinannen': 'hinannen',
'hinsides': 'hinsides',
'hipp': 'hipp',
'hisset': 'hisset',
'hist': 'hist',
'hit': 'hit',
'hitafor': 'hitafor',
'hitenfor': 'hitenfor',
'hitover': 'hitover',
'hittil': 'hittil',
'hjem': 'hjem',
'hjemad': 'hjemad',
'hjemetter': 'hjemetter',
'hjemme': 'hjemme',
'hjemmefra': 'hjemmefra',
'hjemmefra': 'hjemmefra',
'hjemom': 'hjemom',
'hjemover': 'hjemover',
'hk': 'hk',
'hl': 'hl',
'hm': 'hm',
'ho': 'ho',
'hoho': 'hoho',
'hoi': 'hoi',
'holdt': 'holdt',
'hopp': 'hopp',
'hopp sann': 'hopp sann',
'hoppla': 'hoppla',
'hoppsa': 'hoppsa',
'hoppsasa': 'hoppsasa',
'hos': 'hos',
'hosianna': 'hosianna',
'hu': 'hu',
'huff': 'huff',
'hui': 'hui',
'henne': 'hun',
'hun': 'hun',
'hundre': 'hundre',
'hundreogéi': 'hundreogén',
'hundreogén': 'hundreogén',
'hundreogétt': 'hundreogén',
'hurra': 'hurra',
'hurten': 'hurten',
'husimellom': 'husimellom',
'husj': 'husj',
'hut': 'hut',
'huttemegtu': 'huttemegtu',
'huttetu': 'huttetu',
'hva': 'hva',
'hvem': 'hvem',
'hver': 'hver',
'hvert': 'hver',
'hverandre': 'hverandre',
'hvilke': 'hvilken',
'hvilken': 'hvilken',
'hvilket': 'hvilken',
'hvo': 'hvo',
'hvor hen': 'hvor hen',
'hvoretter': 'hvoretter',
'hvorfra': 'hvorfra',
'hvori': 'hvori',
'hvoriblant': 'hvoriblant',
'hvorimot': 'hvorimot',
'hvorledes': 'hvorledes',
'hvormed': 'hvormed',
'hvorom': 'hvorom',
'hvorpå': 'hvorpå',
'hvorvidt': 'hvorvidt',
'hypp': 'hypp',
'hysj': 'hysj',
'hyss': 'hyss',
'hå': 'hå',
'hæ': 'hæ',
'hø': 'hø',
'hør': 'hør',
'i': 'i',
'i aften': 'i aften',
'i aftes': 'i aftes',
'i dag': 'i dag',
'i fall': 'i fall',
'i fjor': 'i fjor',
'i fleng': 'i fleng',
'i forfjor': 'i forfjor',
'i forgårs': 'i forgårs',
'i gjerde': 'i gjerde',
'i gjære': 'i gjære',
'i grunnen': 'i grunnen',
'i går': 'i går',
'i går kveld': 'i går kveld',
'i hel': 'i hel',
'i hende': 'i hende',
'i hjel': 'i hjel',
'i hop': 'i hop',
'i hug': 'i hug',
'i huleste': 'i huleste',
'i hvert fall': 'i hvert fall',
'i koll': 'i koll',
'i kraft': 'i kraft',
'i kveld': 'i kveld',
'i lag': 'i lag',
'i land': 'i land',
'i like måte': 'i like måte',
'i live': 'i live',
'i mak': 'i mak',
'i mente': 'i mente',
'i miss': 'i miss',
'i mist': 'i mist',
'i morgen': 'i morgen',
'i morges': 'i morges',
'i møte': 'i møte',
'i natt': 'i natt',
'i overmorgen': 'i overmorgen',
'i rette': 'i rette',
'i sammen': 'i sammen',
'i senn': 'i senn',
'i sinne': 'i sinne',
'i stand': 'i stand',
'i sted': 'i sted',
'i stedet for': 'i stedet for',
'i stykker': 'i stykker',
'i stå': 'i stå',
'i sund': 'i sund',
'i svime': 'i svime',
'i så fall': 'i så fall',
'i så måte': 'i så måte',
'i søvne': 'i søvne',
'i tide': 'i tide',
'i tilfelle': 'i tilfelle',
'i tu': 'i tu',
'i utide': 'i utide',
'i vare': 'i vare',
'i veg': 'i veg',
'i vei': 'i vei',
'i vente': 'i vente',
'i verk': 'i verk',
'i år': 'i år',
'i. e.': 'i.e.',
'i.e.': 'i.e.',
'ib.': 'ib.',
'ib.': 'ib.',
'ib.': 'ibid.',
'ib.': 'ibid.',
'ibid': 'ibid.',
'ibid.': 'ibid.',
'iblant': 'iblant',
'idet': 'idet',
'ifra': 'ifra',
'ifølge': 'ifølge',
'igjennom': 'igjennom',
'ikring': 'ikring',
'imellom': 'imellom',
'imot': 'imot',
'inga': 'ingen',
'ingen': 'ingen',
'intet': 'ingen',
'ingenting': 'ingenting',
'inkl.': 'inkl.',
'inklusiv': 'inklusiv',
'inklusive': 'inklusive',
'inn': 'inn',
'inn i': 'inn i',
'inn til': 'inn til',
'innad': 'innad',
'innafor': 'innafor',
'innafra': 'innafra',
'innalands': 'innalands',
'innaskjærs': 'innaskjærs',
'inne': 'inne',
'innefra': 'innefra',
'innen': 'innen',
'innenfor': 'innenfor',
'innenfra': 'innenfra',
'innenifra': 'innenifra',
'innenlands': 'innenlands',
'innenskjærs': 'innenskjærs',
'innetter': 'innetter',
'innfor': 'innfor',
'inni': 'inni',
'inniblant': 'inniblant',
'innimellom': 'innimellom',
'innmed': 'innmed',
'innom': 'innom',
'innomhus': 'innomhus',
'innover': 'innover',
'innpå': 'innpå',
'inntil': 'inntil',
'innunder': 'innunder',
'innved': 'innved',
'innåt': 'innåt',
'intet': 'intet',
'isj': 'isj',
'ist': 'ist',
'istedenfor': 'istedenfor',
'istf': 'istf.',
'istf.': 'istf.',
'ja': 'ja',
'ja vel': 'ja vel',
'ja visst': 'ja visst',
'jadden': 'jadden',
'jaggu': 'jaggu',
'jaha': 'jaha',
'jeg': 'jeg',
'meg': 'jeg',
'jemini': 'jemini',
'jf': 'jf.',
'jf.': 'jf.',
'jfr': 'jf.',
'jfr.': 'jf.',
'jvf': 'jf.',
'jvf.': 'jf.',
'jo': 'jo',
'jo visst': 'jo visst',
'joggu': 'joggu',
'joho': 'joho',
'jr': 'jr.',
'jr.': 'jr.',
'jun': 'jr.',
'jun.': 'jr.',
'jun.': 'jun.',
'jur.': 'jur.',
'juris': 'juris',
'jøss': 'jøss',
'jøsses': 'jøsses',
'jøye': 'jøye',
'kB': 'kB',
'kHz': 'kHz',
'kJ': 'kJ',
'kV': 'kV',
'kW': 'kW',
'kWh': 'kWh',
'Kand.': 'kand.',
'kand.': 'kand.',
'kcal': 'kcal',
'kg': 'kg',
'kg.': 'kg',
'kgl': 'kgl.',
'kgl.': 'kgl.',
'kgl res': 'kgl.res.',
'kgl. res.': 'kgl.res.',
'kgl.res.': 'kgl.res.',
'kiss': 'kiss',
'kl': 'kl.',
'kl.': 'kl.',
'klask': 'klask',
'klikk-klakk': 'klikk-klakk',
'kling': 'kling',
'klingeling': 'klingeling',
'kliss-klass': 'kliss-klass',
'klukk': 'klukk',
'km': 'km',
'km.': 'km',
'ko-ko': 'ko-ko',
'konst.': 'konst.',
'kontra': 'kontra',
'kp': 'kp',
'kpm': 'kpm',
'kr': 'kr',
'kr.': 'kr',
'kra': 'kra',
'kring': 'kring',
'kringom': 'kringom',
'kst.': 'kst.',
'kt': 'kt',
'kurr': 'kurr',
'kvakk': 'kvakk',
'kvekk': 'kvekk',
'kvirrevitt': 'kvirrevitt',
'kykeliky': 'kykeliky',
'kynnere': 'kynnere',
'kyrie eleison': 'kyrie eleison',
'L': 'l',
'l': 'l',
'l.': 'l',
'l\'art pour l\'art': 'l\'art pour l\'art',
'lab': 'lab.',
'lab.': 'lab.',
'landimellom': 'landimellom',
'landverts': 'landverts',
'langs': 'langs',
'langsetter': 'langsetter',
'langsmed': 'langsmed',
'langt fra': 'langt fra',
'langt ifra': 'langt ifra',
'langveges': 'langveges',
'langvegesfra': 'langvegesfra',
'langvegs': 'langvegs',
'langvegsfra': 'langvegsfra',
'langveis': 'langveis',
'langveisfra': 'langveisfra',
'lat.': 'lat.',
'laten': 'laten',
'lev vel': 'lev vel',
'Lg': 'lg',
'lg': 'lg',
'lg.': 'lg',
'log': 'lg',
'log.': 'lg',
'lg': 'lg',
'Lic.': 'lic.',
'lic.': 'lic.',
'Lic. Philos.': 'lic.philos.',
'Lic. philos.': 'lic.philos.',
'Lic.philos': 'lic.philos.',
'lic. philos.': 'lic.philos.',
'lic.philos.': 'lic.philos.',
'lign.': 'lign.',
'like ens': 'like ens',
'like til': 'like til',
'likesom': 'likesom',
'lign.': 'likn.',
'likn.': 'likn.',
'liksom': 'liksom',
'lim.': 'lim.',
'lite grann': 'lite grann',
'lm': 'lm',
'ln': 'ln.',
'ln.': 'ln.',
'log': 'log',
'Lim.': 'ltd.',
'Ltd': 'ltd.',
'Ltd.': 'ltd.',
'lim.': 'ltd.',
'ltd': 'ltd.',
'ltd.': 'ltd.',
'lx': 'lx',
'm': 'm',
'm.': 'm',
'm o.h.': 'm o.h.',
'm. o. h.': 'm o.h.',
'm.o.h.': 'm o.h.',
'moh.': 'm o.h.',
'm u.h.': 'm u.h.',
'm. u. h.': 'm u.h.',
'm.u.h.': 'm u.h.',
'muh.': 'm u.h.',
'M-L': 'm-l',
'm-l': 'm-l',
'm. a. o.': 'm.a.o.',
'm.a.o.': 'm.a.o.',
'm m': 'm.m.',
'm. m.': 'm.m.',
'm.m.': 'm.m.',
'm. o. t.': 'm.o.t.',
'm.o.t.': 'm.o.t.',
'm. p. p.': 'm.p.p.',
'm.p.p.': 'm.p.p.',
'mV': 'mV',
'mW': 'mW',
'mag.': 'mag.',
'Mag. Art.': 'mag.art.',
'Mag. art.': 'mag.art.',
'Mag.art': 'mag.art.',
'Mag.art.': 'mag.art.',
'mag. art.': 'mag.art.',
'mag.art': 'mag.art.',
'mag.art.': 'mag.art.',
'Mag. scient': 'mag.scient.',
'Mag.scient': 'mag.scient.',
'mag. scient': 'mag.scient.',
'mag. scient.': 'mag.scient.',
'mag.scient': 'mag.scient.',
'mag.scient.': 'mag.scient.',
'man': 'man',
'mang': 'mang',
'mangt': 'mang',
'mang ei': 'mang en',
'mang en': 'mang en',
'mangt et': 'mang en',
'mange slags': 'mange slags',
'mang en': 'mangt et',
'mannemellom': 'mannemellom',
'marsj': 'marsj',
'mask.': 'mask.',
'mb': 'mb',
'med': 'med',
'med mindre': 'med mindre',
'med.': 'med.',
'medsols': 'medsols',
'medstrøms': 'medstrøms',
'mek': 'mek.',
'mek.': 'mek.',
'mellom': 'mellom',
'men': 'men',
'mens': 'mens',
'm. fl': 'mfl.',
'm. fl.': 'mfl.',
'm.fl': 'mfl.',
'm.fl.': 'mfl.',
'mfl': 'mfl.',
'mfl.': 'mfl.',
'mg': 'mg',
'm. h. t.': 'mht.',
'm.h.t.': 'mht.',
'mht.': 'mht.',
'midtfjords': 'midtfjords',
'midtskips': 'midtskips',
'midtveges': 'midtveges',
'midtvegs': 'midtvegs',
'midtveis': 'midtveis',
'mill': 'mill.',
'mill.': 'mill.',
'mi': 'min',
'min': 'min',
'mine': 'min',
'mitt': 'min',
'mjau': 'mjau',
'ml': 'ml',
'mm': 'mm',
'mon': 'mon',
'mon tro': 'mon tro',
'mon tru': 'mon tru',
'morn': 'morn',
'mot': 'mot',
'mots': 'mots.',
'mots.': 'mots.',
'mrd': 'mrd.',
'mrd.': 'mrd.',
'ms': 'ms',
'mva': 'mva.',
'mva.': 'mva.',
'må hende': 'må hende',
'må vite': 'må vite',
'mø': 'mø',
'nam': 'nam',
'natura': 'natura',
'ned': 'ned',
'ned i': 'ned i',
'ned på': 'ned på',
'nedad': 'nedad',
'nedafor': 'nedafor',
'nedafra': 'nedafra',
'nedatil': 'nedatil',
'nede': 'nede',
'nedefra': 'nedefra',
'nedenfor': 'nedenfor',
'nedenfra': 'nedenfra',
'nedenifra': 'nedenifra',
'nedenom': 'nedenom',
'nedentil': 'nedentil',
'nedenunder': 'nedenunder',
'nedetter': 'nedetter',
'nedi': 'nedi',
'nedigjennom': 'nedigjennom',
'nedimot': 'nedimot',
'nedmed': 'nedmed',
'nedom': 'nedom',
'nedover': 'nedover',
'nedpå': 'nedpå',
'nedstrøms': 'nedstrøms',
'nedved': 'nedved',
'nei': 'nei',
'neste': 'neste',
'ng': 'ng',
'ni': 'ni',
'nitten': 'nitten',
'nitti': 'nitti',
'nja': 'nja',
'NOK': 'nkr',
'Nkr': 'nkr',
'Nkr.': 'nkr',
'nkr': 'nkr',
'nkr.': 'nkr',
'noe': 'noen',
'noen': 'noen',
'nord': 'nord',
'nord om': 'nord om',
'nordafjells': 'nordafjells',
'nordafor': 'nordafor',
'nordafra': 'nordafra',
'nordaust': 'nordaust',
'nordenfjells': 'nordenfjells',
'nordenfor': 'nordenfor',
'nordenfor': 'nordenfor',
'nordenfra': 'nordenfra',
'nordenom': 'nordenom',
'nordetter': 'nordetter',
'nordfra': 'nordfra',
'nordi': 'nordi',
'nordigjennom': 'nordigjennom',
'nordmed': 'nordmed',
'nordom': 'nordom',
'nordost': 'nordost',
'nordover': 'nordover',
'nordpå': 'nordpå',
'nordvest': 'nordvest',
'nordøst': 'nordøst',
'nr': 'nr.',
'nr.': 'nr.',
'null': 'null',
'nåda': 'nåda',
'nærme': 'nærme',
'nøff': 'nøff',
'nøff-nøff': 'nøff-nøff',
'o. a.': 'o.a.',
'o.a.': 'o.a.',
'o. h.': 'o.h.',
'o.h.': 'o.h.',
'o. l.': 'o.l.',
'o. lign.': 'o.l.',
'o. likn': 'o.l.',
'o.l.': 'o.l.',
'o.lign.': 'o.l.',
'o.likn': 'o.l.',
'o. m.': 'o.m.',
'o.m.': 'o.m.',
'o.m.a': 'o.m.a.',
'o.m.a.': 'o.m.a.',
'obs.': 'obs.',
'odont.': 'odont.',
'oecon.': 'oecon.',
'off': 'off.',
'o. fl': 'ofl.',
'o. fl.': 'ofl.',
'o.fl': 'ofl.',
'o.fl.': 'ofl.',
'ofl.': 'ofl.',
'og': 'og',
'oho': 'oho',
'ohoi': 'ohoi',
'oi': 'oi',
'okk': 'okk',
'om': 'om',
'om': 'om',
'om bord': 'om bord',
'om enn': 'om enn',
'om gangen': 'om gangen',
'om kapp': 'om kapp',
'om kull': 'om kull',
'om lag': 'om lag',
'om senn': 'om senn',
'omenskjønt': 'omenskjønt',
'omfram': 'omfram',
'omframt': 'omframt',
'omkr': 'omkr.',
'omkr.': 'omkr.',
'omkring': 'omkring',
'op': 'op.',
'op.': 'op.',
'op. cit.': 'op.cit.',
'op.cit.': 'op.cit.',
'opp': 'opp',
'opp i': 'opp i',
'opp imot': 'opp imot',
'opp mot': 'opp mot',
'opp på': 'opp på',
'opp til': 'opp til',
'oppe': 'oppe',
'oppetter': 'oppetter',
'oppfor': 'oppfor',
'oppi': 'oppi',
'oppigjennom': 'oppigjennom',
'oppimot': 'oppimot',
'oppmed': 'oppmed',
'oppom': 'oppom',
'oppover': 'oppover',
'opptil': 'opptil',
'oppunder': 'oppunder',
'oppved': 'oppved',
'oppå': 'oppå',
'o. s. fr.': 'osfr.',
'o.s.fr.': 'osfr.',
'osfr': 'osfr.',
'osfr.': 'osfr.',
'ost': 'ost',
'o. s. v.': 'osv.',
'o.s.v.': 'osv.',
'osv': 'osv.',
'osv.': 'osv.',
'Ot prp': 'ot.prp.',
'Ot. prp.': 'ot.prp.',
'Ot.prp.': 'ot.prp.',
'ot prp': 'ot.prp.',
'ot. prp.': 'ot.prp.',
'ot.prp.': 'ot.prp.',
'ovabords': 'ovabords',
'ovafor': 'ovafor',
'ovafra': 'ovafra',
'ovapå': 'ovapå',
'ovatil': 'ovatil',
'ovenbords': 'ovenbords',
'ovenfor': 'ovenfor',
'ovenfra': 'ovenfra',
'ovenom': 'ovenom',
'ovenpå': 'ovenpå',
'oventil': 'oventil',
'over': 'over',
'over bord': 'over bord',
'over ende': 'over ende',
'over kors': 'over kors',
'over lang': 'over lang',
'over skrevs': 'over skrevs',
'over styr': 'over styr',
'over tvert': 'over tvert',
'overfor': 'overfor',
'ovf': 'ovf.',
'ovf.': 'ovf.',
'p': 'p',
'p': 'p.',
'p.': 'p.',
'p a': 'p.a.',
'p. a.': 'p.a.',
'p.a.': 'p.a.',
'p. m.': 'p.m.',
'p.m.': 'p.m.',
'p. p.': 'p.p.',
'p.p.': 'p.p.',
'p. p. m.': 'p.p.m.',
'p.p.m.': 'p.p.m.',
'ppm.': 'p.p.m.',
'p r': 'p.r.',
'p. r.': 'p.r.',
'p.r.': 'p.r.',
'p.t.': 'p.t.',
'pa': 'pa.',
'pa.': 'pa.',
'paed.': 'paed.',
'paff': 'paff',
'passee': 'passé',
'passeere': 'passé',
'passeest': 'passé',
'passeeste': 'passé',
'passant': 'passant',
'per': 'per',
'pereat': 'pereat',
'pers': 'pers',
'pf.': 'pf.',
'p. g. a.': 'pga.',
'p.g.a.': 'pga.',
'pga': 'pga.',
'pga.': 'pga.',
'pharm.': 'pharm.',
'philol.': 'philol.',
'philos.': 'philos.',
'piff': 'piff',
'pinadø': 'pinadø',
'pinede': 'pinede',
'pinedød': 'pinedød',
'pip': 'pip',
'piskadausen': 'piskadausen',
'piskadød': 'piskadød',
'pist': 'pist',
'pl': 'pl.',
'pl.': 'pl.',
'plaff': 'plaff',
'plask': 'plask',
'pling': 'pling',
'plopp': 'plopp',
'plump': 'plump',
'polit.': 'polit.',
'post': 'post',
'poste restante': 'poste restante',
'pr': 'pr.',
'pr.': 'pr.',
'prep': 'prep.',
'prep.': 'prep.',
'pret': 'pret.',
'pret.': 'pret.',
'pro': 'pro',
'pron': 'pron.',
'pron.': 'pron.',
'prosit': 'prosit',
'prositt': 'prositt',
'prp': 'prp.',
'prp.': 'prp.',
'pst': 'pst',
'pst': 'pst.',
'pst.': 'pst.',
'psychol.': 'psychol.',
'ptro': 'ptro',
'ptro': 'ptro',
'puff': 'puff',
'puh': 'puh',
'puss': 'puss',
'pvrk': 'pvrk',
'pytt': 'pytt',
'på': 'på',
'pø': 'pø',
'pøh': 'pøh',
'q. e. d.': 'q.e.d.',
'q.e.d.': 'q.e.d.',
'q. s.': 'q.s.',
'q.s.': 'q.s.',
'qua': 'qua',
'quo vadis': 'quo vadis',
'quod erat demonstrandum': 'quod erat demonstrandum',
'rapp': 'rapp',
'ratsj': 'ratsj',
'real.': 'real.',
'rek.': 'rek.',
'renminbi': 'renminbi',
'Res.kap.': 'res.kap.',
'res.kap.': 'res.kap.',
'resp': 'resp.',
'resp.': 'resp.',
'ritsj': 'ritsj',
'rundt': 'rundt',
'rømmen': 'rømmen',
's a': 's.a.',
's. a.': 's.a.',
's.a.': 's.a.',
'S u': 's.u.',
'S. u.': 's.u.',
'S.u.': 's.u.',
'Su.': 's.u.',
's u': 's.u.',
's. u.': 's.u.',
's.u.': 's.u.',
'su.': 's.u.',
's å': 's.å.',
's. å.': 's.å.',
's.å.': 's.å.',
'salvo': 'salvo',
'samme': 'samme',
'sammesteds': 'sammesteds',
'sammestedsfra': 'sammestedsfra',
'samt': 'samt',
'samtlige': 'samtlige',
'sc': 'sc.',
'sc.': 'sc.',
'scient.': 'scient.',
'seg': 'seg',
'sek': 'sek',
'sek.': 'sek',
'seks': 'seks',
'seksten': 'seksten',
'seksti': 'seksti',
'selv': 'selv',
'selv om': 'selv om',
'selveste': 'selveste',
'sen.': 'sen.',
'sr': 'sen.',
'sr.': 'sen.',
'sen.': 'sen.',
'sesam': 'sesam',
'sg': 'sg.',
'sg.': 'sg.',
'sing': 'sg.',
'sing.': 'sg.',
'sg.': 'sg.',
'sidelangs': 'sidelangs',
'sign.': 'sign.',
'simsalabim': 'simsalabim',
'si': 'sin',
'sin': 'sin',
'sine': 'sin',
'sitt': 'sin',
'sing.': 'sing.',
'sing.': 'sing.',
'singott': 'singott',
'sistpå': 'sistpå',
'sju': 'sju',
'sjøl': 'sjøl',
'sjølveste': 'sjølveste',
'skjønt': 'skjønt',
'skål': 'skål',
'slik': 'slik',
'slike': 'slik',
'slikt': 'slik',
'snapp': 'snapp',
'snart': 'snart',
'snipp': 'snipp',
'som': 'som',
'somme': 'somme',
'sommesteds': 'sommesteds',
'somme': 'somt',
'somt': 'somt',
'sr.': 'sr.',
'sr.': 'sr.',
'st': 'st.',
'st.': 'st.',
'St prp': 'st.prp.',
'St. prp': 'st.prp.',
'St. prp.': 'st.prp.',
'St.prp.': 'st.prp.',
'st prp': 'st.prp.',
'st. prp.': 'st.prp.',
'st.prp.': 'st.prp.',
'stk': 'stk.',
'stk.': 'stk.',
'stud.': 'stud.',
'sturten': 'sturten',
'styrbord': 'styrbord',
'svupp': 'svupp',
'syd': 'syd',
'sydetter': 'sydetter',
'sydfra': 'sydfra',
'sydost': 'sydost',
'sydover': 'sydover',
'sydpå': 'sydpå',
'sydvest': 'sydvest',
'sydvest': 'sydvest',
'sytten': 'sytten',
'sytti': 'sytti',
'så framt': 'så framt',
'så fremt': 'så fremt',
'så menn': 'så menn',
'så nær som': 'så nær som',
'så sant': 'så sant',
'så som': 'så som',
'så vel som': 'så vel som',
'så vidt': 'så vidt',
'sådan': 'sådan',
'sådanne': 'sådan',
'sådant': 'sådan',
'såframt': 'såframt',
'såfremt': 'såfremt',
'sånn': 'sånn',
'sånne': 'sånn',
'sånt': 'sånn',
'sønna': 'sønna',
'sønnafjells': 'sønnafjells',
'sønnafor': 'sønnafor',
'sønnafra': 'sønnafra',
'sønnenfjells': 'sønnenfjells',
'sønnenfor': 'sønnenfor',
'sønnenfra': 'sønnenfra',
'sønnenom': 'sønnenom',
'sør': 'sør',
'søraust': 'søraust',
'søren': 'søren',
'søretter': 'søretter',
'sørfra': 'sørfra',
'sørover': 'sørover',
'sørpå': 'sørpå',
'sørvest': 'sørvest',
'sørvest': 'sørvest',
'sørøst': 'sørøst',
'søtten': 'søtten',
't d': 't.d.',
't. d.': 't.d.',
't.d.': 't.d.',
't. o. m.': 't.o.m.',
't.o.m.': 't.o.m.',
'tbc': 'tbc',
'ti': 'ti',
'tikk': 'tikk',
'tikk-takk': 'tikk-takk',
'til': 'til',
'til dels': 'til dels',
'til dømes': 'til dømes',
'til fals': 'til fals',
'til like': 'til like',
'til live': 'til live',
'til livs': 'til livs',
'til orde': 'til orde',
'til overs': 'til overs',
'til stede': 'til stede',
'tilbake': 'tilbake',
'tillikemed': 'tillikemed',
'tja': 'tja',
'tjo': 'tjo',
'tjue': 'tjue',
'to': 'to',
'tolv': 'tolv',
'tralala': 'tralala',
'trass': 'trass',
'tre': 'tre',
'tretten': 'tretten',
'tretti': 'tretti',
'triplo': 'triplo',
'tripp': 'tripp',
'trommelom': 'trommelom',
'tross': 'tross',
'tupp': 'tupp',
'tusen': 'tusen',
'tvi': 'tvi',
'tyss': 'tyss',
'tøff': 'tøff',
'uaktet': 'uaktet',
'uansett': 'uansett',
'uff': 'uff',
'ult.': 'ult.',
'ultimo': 'ultimo',
'undas': 'undas',
'under': 'under',
'under vær': 'under vær',
'undervegs': 'undervegs',
'underveis': 'underveis',
'unna': 'unna',
'unntagen': 'unntagen',
'unntatt': 'unntatt',
'ut': 'ut',
'ut av': 'ut av',
'ut i': 'ut i',
'ut over': 'ut over',
'ut på': 'ut på',
'utad': 'utad',
'utafor': 'utafor',
'utafra': 'utafra',
'utalands': 'utalands',
'utapå': 'utapå',
'utav': 'utav',
'ute': 'ute',
'utefra': 'utefra',
'uten': 'uten',
'utenfor': 'utenfor',
'utenfra': 'utenfra',
'utenlands': 'utenlands',
'utenom': 'utenom',
'utenpå': 'utenpå',
'utetter': 'utetter',
'utfor': 'utfor',
'uti': 'uti',
'utmed': 'utmed',
'utom': 'utom',
'utomhus': 'utomhus',
'utover': 'utover',
'utpå': 'utpå',
'utved': 'utved',
'v': 'v.',
'v.': 'v.',
'vn.': 'v.',
'van': 'van',
'varsku': 'varsku',
've': 've',
'ved': 'ved',
'ved like': 'ved like',
'veggimellom': 'veggimellom',
'vekk': 'vekk',
'vekke': 'vekke',
'verken': 'verken',
'versus': 'versus',
'vest': 'vest',
'vesta': 'vesta',
'vestafjells': 'vestafjells',
'vestafor': 'vestafor',
'vestafra': 'vestafra',
'vestenfjells': 'vestenfjells',
'vestenfor': 'vestenfor',
'vestenfra': 'vestenfra',
'vestenom': 'vestenom',
'vestfra': 'vestfra',
'vestover': 'vestover',
'vestpå': 'vestpå',
'vet.': 'vet.',
'oss': 'vi',
'vi': 'vi',
'via': 'via',
'vips': 'vips',
'vis-à-vis': 'vis-à-vis',
'Vit. Ass.': 'vit.ass.',
'Vit. ass.': 'vit.ass.',
'vit. ass.': 'vit.ass.',
'vit.ass.': 'vit.ass.',
'vivat': 'vivat',
'voff': 'voff',
'voff-voff': 'voff-voff',
'von': 'von',
'vov': 'vov',
'vov-vov': 'vov-vov',
'v. s. a.': 'vsa.',
'v.s.a.': 'vsa.',
'vsa': 'vsa.',
'vsa.': 'vsa.',
'vår': 'vår',
'våre': 'vår',
'vårt': 'vår',
'yd': 'yd',
'zero': 'zero',
'å': 'å',
'å': 'å',
'åhå': 'åhå',
'åt': 'åt',
'åtte': 'åtte',
'åtti': 'åtti',
'æ': 'æ',
'æsj': 'æsj',
'ørten': 'ørten',
'øst': 'øst',
'østa': 'østa',
'østafjells': 'østafjells',
'østafor': 'østafor',
'østafra': 'østafra',
'østenfjells': 'østenfjells',
'østenfor': 'østenfor',
'østenfra': 'østenfra',
'østenom': 'østenom',
'østetter': 'østetter',
'østetter': 'østetter',
'østfra': 'østfra',
'østfra': 'østfra',
'østover': 'østover',
'østover': 'østover',
'østpå': 'østpå',
'østpå': 'østpå',
'egen': 'egen',
'eget': 'egen',
'egne': 'egen',
'fersken': 'fersken',
'hvis': 'hvis',
'hvis': 'hvis',
'den': 'den',
'det': 'det',
'eder': 'eder',
'enøk': 'enøk',
'frå': 'frå',
'frå': 'frå',
'hundre og éi': 'hundre og én',
'hundre og én': 'hundre og én',
'hundre og étt': 'hundre og én',
'hverken': 'hverken',
'ifrå': 'ifrå',
'i mellomtida': 'i mellomtida',
'i mellomtiden': 'i mellomtiden',
'kille': 'kille',
'min': 'min',
'nere': 'nere',
'nerimot': 'nerimot',
'nere': 'nere',
'pang': 'pang',
'pardong': 'pardong',
'pro': 'pro',
'pro': 'pro',
'restante': 'restante',
'sin': 'sin',
'sjøl om': 'sjøl om',
'svibrent': 'svibrent',
'tan': 'tan',
'tjueei': 'tjueen',
'tjueen': 'tjueen',
'tjueett': 'tjueen',
'trapp': 'trapp',
'tredve': 'tredve',
'trom': 'trom',
'tvibrent': 'tvibrent',
'tyve': 'tyve',
'UHF': 'UHF',
'vilden': 'vilden',
'villen': 'villen',
'villen': 'villen',
'vilden': 'vilden',
'by': 'by',
'de': 'de',
'efter': 'efter',
'et': 'et',
'fyterakkern': 'fyterakkern',
'hi': 'hi',
'ho': 'ho',
'ho': 'ho',
'viss': 'viss',
'kynnerer': 'kynnerer',
'ner': 'ner',
'ner': 'ner',
'nå': 'nå',
'syv': 'syv',
'viss': 'viss',
'frå borde': 'frå borde',
'frå borde': 'frå borde',
'til og med': 'til og med',
'av': 'av',
'c': 'c',
'CD': 'CD',
'E': 'E',
'f': 'f',
'g': 'g',
'g.': 'g',
'gt': 'g',
'gt.': 'g',
'Mrs.': 'Mrs.',
'mrs.': 'Mrs.',
't': 't',
't.': 't',
'U': 'U',
'u': 'U',
'u': 'u',
'v': 'v',
'X': 'X',
'YAP': 'YAP',
'g': 'g',
'gr': 'g',
'gr.': 'g',
's': 's',
'hvems': 'hvems',
'i hu': 'i hu',
'd. s.': 'd.s.',
'd.s.': 'd.s.',
'm. v.': 'mv.',
'm.v.': 'mv.',
'mv': 'mv.',
'mv.': 'mv.',
'n': 'n.',
'n.': 'n.',
'dim.': 'dim.',
'ad-': 'ad-',
'kvitt': 'kvitt',
'innendørs': 'innendørs',
'utendørs': 'utendørs',
'siden': 'siden',
'C.': 'C.',
'M.': 'M.',
'noen': 'noen',
'noen': 'noen',
'begge': 'begge',
'enhver': 'enhver',
'ingen': 'ingen',
'ingen': 'ingen',
'intet': 'intet',
'hvilken': 'hvilken',
'somt': 'somt',
'somme': 'somme',
'u-': 'u-',
'ASA': 'ASA',
'laden': 'laden',
'varsko': 'varsko',
'ennskjønt': 'ennskjønt',
'omennskjønt': 'omennskjønt',
'jepp': 'jepp'
}
